#!/usr/bin/env python3

"""Module containing the Canion class and the command line interface."""
import os
import zipfile
from typing import Optional
from pathlib import Path
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class Canion(BiobbObject):
    """
    | biobb_dna Canion
    | Wrapper for the Canion executable  that is part of the Curves+ software suite.
    | Analyzes the trajectory of ions around a DNA molecule.

    Args:
        input_cdi_path (str): Trajectory input file. File type: input. `Sample file <https://mmb.irbbarcelona.org/biobb-dev/biobb-api/public/samples/THGA_K.cdi>`_. Accepted formats: cdi (edam:format_2330).
        input_afr_path (str): Helical axis frames corresponding to the input conformation to be analyzed. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/THGA.afr>`_. Accepted formats: afr (edam:format_2330).
        input_avg_struc_path (str): Average DNA conformation. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/THGA_avg.pdb>`_. Accepted formats: pdb (edam:format_1476).
        output_zip_path (str): Filename for .zip files containing Canion output files. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/canion_output.zip>`_. Accepted formats: zip (edam:format_3987).
        properties (dict):
            * **bases** (*str*) - (None) Sequence of bases to be analyzed (default is blank, meaning no specified sequence).
            * **type** (*str*) - ('*') Ions (or atoms) to be analyzed. Options are 'Na+', 'K', 'K+', 'Cl', 'Cl-', 'CL', 'P', 'C1*', 'NH1', 'NH2', 'NZ', '1' for all cations, '-1' for all anions, '0' for neutral species or '*' for all available data.
            * **dlow** (*float*) - (0) Select starting segment of the oglimer to analyze. If both dhig and dlow are 0, entire oglimer is analyzed.
            * **dhig** (*float*) - (0) Select ending segment of the oglimer to analyze, being the maximum value the total number of base pairs in the oligomer. If both dhig and dlow are 0, entire oglimer is analyzed.
            * **rlow** (*float*) - (0) Minimal distances from the helical axis taken into account in the analysis.
            * **rhig** (*float*) - (0) Maximal distances from the helical axis taken into account in the analysis.
            * **alow** (*float*) - (0) Minimal angle range to analyze.
            * **ahig** (*float*) - (360) Maximal angle range to analyze.
            * **itst** (*int*) - (0) Number of first snapshot to be analyzed.
            * **itnd** (*int*) - (0) Number of last snapshot to be analyzed.
            * **itdel** (*int*) - (1) Spacing between analyzed snapshots.
            * **rmsf** (*bool*) - (False) If set to True uses the combination of the helical ion parameters and an average helical axis to map the ions into Cartesian space and then calculates their average position (pdb output) and their root mean square fluctuation values (rmsf output). A single pass rmsf algorithm to make this calculation possible with a single read of the trajectory file. This option is generally used for solute atoms and not for solvent molecules or ions.
            * **circ** (*bool*) - (False) If set to True, minicircles are analyzed.
            * **binary_path** (*str*) - (Canion) Path to Canion executable, otherwise the program wil look for Canion executable in the binaries folder.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.curvesplus.biobb_canion import biobb_canion
            prop = {
                'type': 'K+',
                'bases': 'G'
            }
            biobb_canion(
                input_cdi_path='/path/to/input.cdi',
                input_afr_path='/path/to/input.afr',
                input_avg_struc_path='/path/to/input.pdb',
                output_zip_path='/path/to/output.zip',
                properties=prop)
    Info:
        * wrapped_software:
            * name: Canion
            * version: >=2.6
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
            self, input_cdi_path, input_afr_path, input_avg_struc_path,
            output_zip_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_cdi_path': input_cdi_path,
                'input_afr_path': input_afr_path,
                'input_avg_struc_path': input_avg_struc_path,
            },
            'out': {
                'output_zip_path': output_zip_path
            }
        }

        # Properties specific for BB
        self.binary_path = properties.get('binary_path', 'Canion')
        self.bases = properties.get('bases', None)
        self.type = properties.get('type', '*')
        self.dlow = properties.get('dlow', 0)
        self.dhig = properties.get('dhig', 0)
        self.rlow = properties.get('rlow', 0)
        self.rhig = properties.get('rhig', 0)
        self.alow = properties.get('alow', 0)
        self.ahig = properties.get('ahig', 360)
        self.itst = properties.get('itst', 0)
        self.itnd = properties.get('itnd', 0)
        self.itdel = properties.get('itdel', 1)
        self.rmsf = ".t." if properties.get('rmsf', False) else ".f."
        self.circ = ".t." if properties.get('circ', False) else ".f."
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`Canion <biobb_dna.curvesplus.biobb_canion.Canion>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        ion_type_options = [
            'Na+',
            'K',
            'K+',
            'Cl',
            'Cl-',
            'CL',
            'P',
            'C1*',
            'NH1',
            'NH2',
            'NZ',
            '1',
            '-1',
            '0',
            '*'
        ]
        if self.type not in ion_type_options:
            raise ValueError(("Invalid value for property type! "
                              f"Option include: {ion_type_options}"))

        # define temporary file names
        input_cdi_file = Path(self.stage_io_dict['in']['input_cdi_path']).name
        input_afr_file = Path(self.stage_io_dict['in']['input_afr_path']).name
        input_avg_struc = Path(self.stage_io_dict['in']['input_avg_struc_path']).name

        # change directory to temporary folder
        original_directory = os.getcwd()
        os.chdir(self.stage_io_dict.get("unique_dir", ""))

        # create intructions
        instructions = [
            f"{self.binary_path} <<! ",
            "&inp",
            "  lis=canion_output,",
            f"  dat={input_cdi_file[:-4]},",
            f"  axfrm={input_afr_file[:-4]},",
            f"  solute={input_avg_struc[:-4]},",
            f"  type={self.type},",
            f"  dlow={self.dlow},",
            f"  dhig={self.dhig},",
            f"  rlow={self.rlow},",
            f"  rhig={self.rhig},",
            f"  alow={self.alow},",
            f"  ahig={self.ahig},",
            f"  itst={self.itst},",
            f"  itnd={self.itnd},",
            f"  itdel={self.itdel},",
            f"  rmsf={self.rmsf},",
            f"  circ={self.circ},"]
        if self.bases is not None:
            # add topology file if needed
            fu.log('Appending sequence of bases to be searched to command',
                   self.out_log, self.global_log)
            instructions.append(f"  seq={self.bases},")
        instructions.append("&end")
        instructions.append("!")
        self.cmd = ["\n".join(instructions)]

        fu.log('Creating command line with instructions and required arguments',
               self.out_log, self.global_log)
        # Run Biobb block
        self.run_biobb()

        # change back to original directory
        os.chdir(original_directory)

        # create zipfile and write output inside
        zf = zipfile.ZipFile(
            Path(self.stage_io_dict["out"]["output_zip_path"]),
            "w")
        for curves_outfile in Path(self.stage_io_dict.get("unique_dir", "")).glob("canion_output*"):
            if curves_outfile.suffix not in (".zip"):
                zf.write(curves_outfile, arcname=curves_outfile.name)
        zf.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def biobb_canion(
        input_cdi_path: str, input_afr_path: str, input_avg_struc_path: str,
        output_zip_path: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`Canion <biobb_dna.curvesplus.biobb_canion.Canion>` class and
    execute the :meth:`launch() <biobb_dna.curvesplus.biobb_canion.Canion.launch>` method."""
    return Canion(**dict(locals())).launch()


biobb_canion.__doc__ = Canion.__doc__
main = Canion.get_main(biobb_canion, "Execute Canion form the Curves+ software suite.")

if __name__ == '__main__':
    main()



def inverse_complement(sequence, dna=True):
    """compute inverse complement sequence."""
    if dna:
        A_complement = "T"
    else:
        A_complement = "U"
    complement = {
        "A": A_complement,
        A_complement: "A",
        "G": "C",
        "C": "G"
    }
    return "".join([complement[b] for b in sequence])[::-1]



#!/usr/bin/env python3

"""Module containing the IntraHelParCorrelation class and the command line interface."""
from typing import Optional

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger
from biobb_dna.utils.loader import load_data


class IntraHelParCorrelation(BiobbObject):
    """
    | biobb_dna IntraHelParCorrelation
    | Calculate correlation between helical parameters for a single intra-base pair.
    | Calculate correlation between helical parameters for a single intra-base pair.

    Args:
        input_filename_shear (str): Path to .csv file with data for helical parameter 'shear'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_shear_A.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_stretch (str): Path to .csv file with data for helical parameter 'stretch'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_stretch_A.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_stagger (str): Path to .csv file with data for helical parameter 'stagger'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_stagger_A.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_buckle (str): Path to .csv file with data for helical parameter 'buckle'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_buckle_A.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_propel (str): Path to .csv file with data for helical parameter 'propeller'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_propel_A.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_opening (str): Path to .csv file with data for helical parameter 'opening'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/series_opening_A.csv>`_. Accepted formats: csv (edam:format_3752).
        output_csv_path (str): Path to directory where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/intra_hpcorr_ref.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/intra_hpcorr_ref.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **base** (*str*) - (None) Name of base analyzed.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.intrabp_correlations.intrahpcorr import intrahpcorr

            prop = {
                'base': 'A',
            }
            intrahpcorr(
                input_filename_shear='path/to/shear.csv',
                input_filename_stretch='path/to/stretch.csv',
                input_filename_stagger='path/to/stagger.csv',
                input_filename_buckle='path/to/buckle.csv',
                input_filename_propel='path/to/propel.csv',
                input_filename_opening='path/to/opening.csv',
                output_csv_path='path/to/output/file.csv',
                output_jpg_path='path/to/output/file.jpg',
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
            self, input_filename_shear, input_filename_stretch,
            input_filename_stagger, input_filename_buckle,
            input_filename_propel, input_filename_opening,
            output_csv_path, output_jpg_path,
            properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_filename_shear': input_filename_shear,
                'input_filename_stretch': input_filename_stretch,
                'input_filename_stagger': input_filename_stagger,
                'input_filename_buckle': input_filename_buckle,
                'input_filename_propel': input_filename_propel,
                'input_filename_opening': input_filename_opening
            },
            'out': {
                'output_csv_path': output_csv_path,
                'output_jpg_path': output_jpg_path
            }
        }

        self.properties = properties
        self.base = properties.get("base", None)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`IntraHelParCorrelation <intrabp_correlations.intrahpcorr.IntraHelParCorrelation>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # read input
        shear = load_data(self.stage_io_dict["in"]["input_filename_shear"])
        stretch = load_data(self.stage_io_dict["in"]["input_filename_stretch"])
        stagger = load_data(self.stage_io_dict["in"]["input_filename_stagger"])
        buckle = load_data(self.stage_io_dict["in"]["input_filename_buckle"])
        propel = load_data(self.stage_io_dict["in"]["input_filename_propel"])
        opening = load_data(self.stage_io_dict["in"]["input_filename_opening"])

        # get base
        if self.base is None:
            self.base = shear.columns[0]

        # make matrix
        # coordinates = ["shear", "stretch", "stagger", "buckle", "propel", "opening"]
        coordinates = [
            "shear", "stretch", "stagger", "buckle", "propel", "opening"]
        corr_matrix = pd.DataFrame(
            np.eye(6, 6), index=coordinates, columns=coordinates)

        # shear
        # corr_matrix["shear"]["stretch"] = shear.corrwith(stretch, method="pearson")
        corr_matrix.loc["stretch", "shear"] = shear.corrwith(stretch, method="pearson").values[0]
        # corr_matrix["shear"]["stagger"] = shear.corrwith(stagger, method="pearson")
        corr_matrix.loc["stagger", "shear"] = shear.corrwith(stagger, method="pearson").values[0]
        # corr_matrix["shear"]["buckle"] = shear.corrwith(buckle, method=self.circlineal)
        corr_matrix.loc["buckle", "shear"] = shear.corrwith(buckle, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["shear"]["propel"] = shear.corrwith(propel, method=self.circlineal)
        corr_matrix.loc["propel", "shear"] = shear.corrwith(propel, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["shear"]["opening"] = shear.corrwith(opening, method=self.circlineal)
        corr_matrix.loc["opening", "shear"] = shear.corrwith(opening, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["stretch"]["shear"] = corr_matrix["shear"]["stretch"]
        corr_matrix.loc["shear", "stretch"] = corr_matrix.loc["stretch", "shear"]
        # corr_matrix["stagger"]["shear"] = corr_matrix["shear"]["stagger"]
        corr_matrix.loc["shear", "stagger"] = corr_matrix.loc["stagger", "shear"]
        # corr_matrix["buckle"]["shear"] = corr_matrix["shear"]["buckle"]
        corr_matrix.loc["shear", "buckle"] = corr_matrix.loc["buckle", "shear"]
        # corr_matrix["propel"]["shear"] = corr_matrix["shear"]["propel"]
        corr_matrix.loc["shear", "propel"] = corr_matrix.loc["propel", "shear"]
        # corr_matrix["opening"]["shear"] = corr_matrix["shear"]["opening"]
        corr_matrix.loc["shear", "opening"] = corr_matrix.loc["opening", "shear"]

        # stretch
        # corr_matrix["stretch"]["stagger"] = stretch.corrwith(stagger, method="pearson")
        corr_matrix.loc["stagger", "stretch"] = stretch.corrwith(stagger, method="pearson").values[0]
        # corr_matrix["stretch"]["buckle"] = stretch.corrwith(buckle, method=self.circlineal)
        corr_matrix.loc["buckle", "stretch"] = stretch.corrwith(buckle, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["stretch"]["propel"] = stretch.corrwith(propel, method=self.circlineal)
        corr_matrix.loc["propel", "stretch"] = stretch.corrwith(propel, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["stretch"]["opening"] = stretch.corrwith(opening, method=self.circlineal)
        corr_matrix.loc["opening", "stretch"] = stretch.corrwith(opening, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["stagger"]["stretch"] = corr_matrix["stretch"]["stagger"]
        corr_matrix.loc["stretch", "stagger"] = corr_matrix.loc["stagger", "stretch"]
        # corr_matrix["buckle"]["stretch"] = corr_matrix["stretch"]["buckle"]
        corr_matrix.loc["stretch", "buckle"] = corr_matrix.loc["buckle", "stretch"]
        # corr_matrix["propel"]["stretch"] = corr_matrix["stretch"]["propel"]
        corr_matrix.loc["stretch", "propel"] = corr_matrix.loc["propel", "stretch"]
        # corr_matrix["opening"]["stretch"] = corr_matrix["stretch"]["opening"]
        corr_matrix.loc["stretch", "opening"] = corr_matrix.loc["opening", "stretch"]

        # stagger
        # corr_matrix["stagger"]["buckle"] = stagger.corrwith(buckle, method=self.circlineal)
        corr_matrix.loc["buckle", "stagger"] = stagger.corrwith(buckle, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["stagger"]["propel"] = stagger.corrwith(propel, method=self.circlineal)
        corr_matrix.loc["propel", "stagger"] = stagger.corrwith(propel, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["stagger"]["opening"] = stagger.corrwith(opening, method=self.circlineal)
        corr_matrix.loc["opening", "stagger"] = stagger.corrwith(opening, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["buckle"]["stagger"] = corr_matrix["stagger"]["buckle"]
        corr_matrix.loc["stagger", "buckle"] = corr_matrix.loc["buckle", "stagger"]
        # corr_matrix["propel"]["stagger"] = corr_matrix["stagger"]["propel"]
        corr_matrix.loc["stagger", "propel"] = corr_matrix.loc["propel", "stagger"]
        # corr_matrix["opening"]["stagger"] = corr_matrix["stagger"]["opening"]
        corr_matrix.loc["stagger", "opening"] = corr_matrix.loc["opening", "stagger"]

        # buckle
        # corr_matrix["buckle"]["propel"] = buckle.corrwith(propel, method=self.circular)
        corr_matrix.loc["propel", "buckle"] = buckle.corrwith(propel, method=self.circular).values[0]  # type: ignore
        # corr_matrix["buckle"]["opening"] = buckle.corrwith(opening, method=self.circular)
        corr_matrix.loc["opening", "buckle"] = buckle.corrwith(opening, method=self.circular).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["propel"]["buckle"] = corr_matrix["buckle"]["propel"]
        corr_matrix.loc["buckle", "propel"] = corr_matrix.loc["propel", "buckle"]
        # corr_matrix["opening"]["buckle"] = corr_matrix["buckle"]["opening"]
        corr_matrix.loc["buckle", "opening"] = corr_matrix.loc["opening", "buckle"]

        # propel
        # corr_matrix["propel"]["opening"] = propel.corrwith(opening, method=self.circular)
        corr_matrix.loc["opening", "propel"] = propel.corrwith(opening, method=self.circular).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["opening"]["propel"] = corr_matrix["propel"]["opening"]
        corr_matrix.loc["propel", "opening"] = corr_matrix.loc["opening", "propel"]

        # save csv data
        corr_matrix.to_csv(self.stage_io_dict["out"]["output_csv_path"])

        # create heatmap
        fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
        axs.pcolor(corr_matrix)
        # Loop over data dimensions and create text annotations.
        for i in range(len(corr_matrix)):
            for j in range(len(corr_matrix)):
                axs.text(
                    j+.5,
                    i+.5,
                    f"{corr_matrix[coordinates[j]].loc[coordinates[i]]:.2f}",
                    ha="center",
                    va="center",
                    color="w")
        axs.set_xticks([i + 0.5 for i in range(len(corr_matrix))])
        axs.set_xticklabels(corr_matrix.columns, rotation=90)
        axs.set_yticks([i+0.5 for i in range(len(corr_matrix))])
        axs.set_yticklabels(corr_matrix.index)
        axs.set_title(
            "Helical Parameter Correlation "
            f"for Base Pair Step \'{self.base}\'")
        fig.tight_layout()
        fig.savefig(
            self.stage_io_dict['out']['output_jpg_path'],
            format="jpg")
        plt.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0

    def get_corr_method(self, corrtype1, corrtype2):
        if corrtype1 == "circular" and corrtype2 == "linear":
            method = self.circlineal
        if corrtype1 == "linear" and corrtype2 == "circular":
            method = self.circlineal
        elif corrtype1 == "circular" and corrtype2 == "circular":
            method = self.circular
        else:
            method = "pearson"
        return method

    @staticmethod
    def circular(x1, x2):
        x1 = x1 * np.pi / 180
        x2 = x2 * np.pi / 180
        diff_1 = np.sin(x1 - x1.mean())
        diff_2 = np.sin(x2 - x2.mean())
        num = (diff_1 * diff_2).sum()
        den = np.sqrt((diff_1 ** 2).sum() * (diff_2 ** 2).sum())
        return num / den

    @staticmethod
    def circlineal(x1, x2):
        x2 = x2 * np.pi / 180
        rc = np.corrcoef(x1, np.cos(x2))[1, 0]
        rs = np.corrcoef(x1, np.sin(x2))[1, 0]
        rcs = np.corrcoef(np.sin(x2), np.cos(x2))[1, 0]
        num = (rc ** 2) + (rs ** 2) - 2 * rc * rs * rcs
        den = 1 - (rcs ** 2)
        correlation = np.sqrt(num / den)
        if np.corrcoef(x1, x2)[1, 0] < 0:
            correlation *= -1
        return correlation


def intrahpcorr(
        input_filename_shear: str, input_filename_stretch: str,
        input_filename_stagger: str, input_filename_buckle: str,
        input_filename_propel: str, input_filename_opening: str,
        output_csv_path: str, output_jpg_path: str,
        properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`IntraHelParCorrelation <intrabp_correlations.intrahpcorr.IntraHelParCorrelation>` class and
    execute the :meth:`launch() <intrabp_correlations.intrahpcorr.IntraHelParCorrelation.launch>` method."""
    return IntraHelParCorrelation(**dict(locals())).launch()


intrahpcorr.__doc__ = IntraHelParCorrelation.__doc__
main = IntraHelParCorrelation.get_main(intrahpcorr, "Load helical parameter file and save base data individually.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the HelParBimodality class and the command line interface."""
import os
import zipfile
from typing import Optional
from pathlib import Path

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture  # type: ignore
from biobb_dna.utils import constants
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger
from biobb_dna.utils.loader import load_data


class HelParBimodality(BiobbObject):
    """
    | biobb_dna HelParBimodality
    | Determine binormality/bimodality from a helical parameter series dataset.
    | Determine binormality/bimodality from a helical parameter series dataset.

    Args:
        input_csv_file (str): Path to .csv file with helical parameter series. If `input_zip_file` is passed, this should be just the filename of the .csv file inside .zip.  File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/dna/series_shift_AT.csv>`_. Accepted formats: csv (edam:format_3752).
        input_zip_file (str) (Optional): .zip file containing the `input_csv_file` .csv file. File type: input. Accepted formats: zip (edam:format_3987).
        output_csv_path (str): Path to .csv file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/AT_shift_bimod.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/AT_shift_bimod.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **helpar_name** (*str*) - (Optional) helical parameter name.
            * **confidence_level** (*float*) - (5.0) Confidence level for Byes Factor test (in percentage).
            * **max_iter** (*int*) - (400) Number of maximum iterations for EM algorithm.
            * **tol** (*float*) - (1e-5) Tolerance value for EM algorithm.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.1

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.dna.dna_bimodality import dna_bimodality

            prop = {
                'max_iter': 500,
            }
            dna_bimodality(
                input_csv_file='filename.csv',
                input_zip_file='/path/to/input.zip',
                output_csv_path='/path/to/output.csv',
                output_jpg_path='/path/to/output.jpg',
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(self, input_csv_file, output_csv_path,
                 output_jpg_path, input_zip_file=None,
                 properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_csv_file': input_csv_file,
                'input_zip_file': input_zip_file
            },
            'out': {
                'output_csv_path': output_csv_path,
                'output_jpg_path': output_jpg_path
            }
        }

        # Properties specific for BB
        self.confidence_level = properties.get(
            "confidence_level", 5.0)
        self.max_iter = properties.get(
            "max_iter", 400)
        self.tol = properties.get(
            "tol", 1e-5)
        self.helpar_name = properties.get("helpar_name", None)
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`HelParBimodality <dna.dna_bimodality.HelParBimodality>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # get helical parameter from filename if not specified
        if self.helpar_name is None:
            for hp in constants.helical_parameters:
                if self.stage_io_dict.get("in", {}).get("input_zip_file") is not None:
                    condition_2 = (
                        hp.lower() in Path(self.stage_io_dict['in']['input_zip_file']).name.lower())
                else:
                    condition_2 = False
                condition_1 = hp.lower() in Path(
                    self.stage_io_dict['in']['input_csv_file']).name.lower()
                if (condition_1 or condition_2):
                    self.helpar_name = hp
            if self.helpar_name is None:
                raise ValueError(
                    "Helical parameter name can't be inferred from file, "
                    "so it must be specified!")
        else:
            if self.helpar_name not in constants.helical_parameters:
                raise ValueError(
                    "Helical parameter name is invalid! "
                    f"Options: {constants.helical_parameters}")

        # get  unit from helical parameter name
        if self.helpar_name in constants.hp_angular:
            self.hp_unit = "Degrees"
        else:
            self.hp_unit = "Angstroms"

        # resolve output
        output_csv_path = Path(
            self.stage_io_dict['out']['output_csv_path']).resolve()
        output_jpg_path = Path(
            self.stage_io_dict['out']['output_jpg_path']).resolve()

        # change directory to temporary folder
        original_directory = os.getcwd()
        os.chdir(self.stage_io_dict.get("unique_dir", ""))

        # read input
        if self.stage_io_dict.get("in", {}).get("input_zip_file") is not None:
            # if zipfile is specified, extract to temporary folder
            with zipfile.ZipFile(
                    self.stage_io_dict['in']['input_zip_file'],
                    'r') as zip_ref:
                zip_ref.extractall(self.stage_io_dict.get("unique_dir", ""))

        data = load_data(Path(self.stage_io_dict['in']['input_csv_file']).name)

        means, variances, bics, weights = self.fit_to_model(data)
        uninormal, binormal, insuf_ev = self.bayes_factor_criteria(
            bics[0], bics[1])

        if binormal:
            maxm = np.argmax(means[1])
            minm = np.argmin(means[1])
            mean1 = means[1][minm]
            var1 = variances[1][minm]
            w1 = weights[1][minm]
            mean2 = means[1][maxm]
            var2 = variances[1][maxm]
            w2 = weights[1][maxm]
            bimodal = self.helguero_theorem(mean1, mean2, var1, var2)
        else:
            mean1 = means[0][0]
            var1 = variances[0][0]
            w1 = weights[0][0]
            mean2, var2, w2 = np.nan, np.nan, 0
            bimodal = False
        info = dict(
            binormal=binormal,
            uninormal=uninormal,
            insuf_ev=insuf_ev,
            bimodal=bimodal,
            mean1=mean1,
            mean2=mean2,
            var1=var1,
            var2=var2,
            w1=w1,
            w2=w2)

        # save tables
        pd.DataFrame(info, index=data.columns).to_csv(output_csv_path)

        # make and save plot
        data_size = len(data)
        synth1 = np.random.normal(
            loc=info['mean1'],
            scale=np.sqrt(info['var1']),
            size=int(data_size * info['w1']))
        synth2 = np.random.normal(
            loc=info['mean2'],
            scale=np.sqrt(info['var2']),
            size=int(data_size * info['w2']))

        plt.figure()
        alpha = 0.7
        bins = 100
        if binormal:
            label1 = "Low State"
        else:
            label1 = "Single State"
        out = plt.hist(
            synth1, bins=bins, alpha=alpha, density=True, label=label1)
        ylim = max(out[0])  # type: ignore
        plt.vlines(info['mean1'], 0, ylim, colors="r", linestyles="dashed")
        if binormal:
            out = plt.hist(
                synth2, bins=bins, alpha=alpha, density=True, label="high state")
            ylim = max(out[0])  # type: ignore
            plt.vlines(info['mean2'], 0, ylim, colors="r", linestyles="dashed")
        plt.legend()
        plt.ylabel("Density")
        plt.xlabel(f"{self.helpar_name.capitalize()} ({self.hp_unit})")
        plt.title(f"Distribution of {self.helpar_name} states")
        plt.savefig(output_jpg_path, format="jpg")
        plt.close()

        # change back to original directory
        os.chdir(original_directory)

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0

    def fit_to_model(self, data):
        """
        Fit data to Gaussian Mixture models.
        Return dictionary with distribution data.
        """
        means = []
        variances = []
        bics = []
        weights = []
        for n_components in (1, 2):
            gmm = GaussianMixture(
                n_components=n_components,
                max_iter=self.max_iter,
                tol=self.tol)
            gmm = gmm.fit(data)
            m = gmm.means_.flatten()  # type: ignore
            v = gmm.covariances_.flatten()  # type: ignore
            b = gmm.bic(data)
            w = gmm.weights_.flatten()  # type: ignore
            means.append(m)
            variances.append(v)
            bics.append(b)
            weights.append(w)
        return means, variances, bics, weights

    def bayes_factor_criteria(self, bic1, bic2):
        diff_bic = bic2 - bic1
        # probability of a two-component model
        p = 1 / (1 + np.exp(0.5*diff_bic))
        if p == np.nan:
            if bic1 == np.nan:
                p = 1
            elif bic2 == np.nan:
                p = 0

        uninormal = p < (self.confidence_level / 100)
        binormal = p > (1 - (self.confidence_level / 100))
        insuf_ev = True if (not uninormal and not binormal) else False
        return uninormal, binormal, insuf_ev

    def helguero_theorem(self, mean1, mean2, var1, var2):
        r = var1 / var2
        separation_factor = np.sqrt(
            -2 + 3*r + 3*r**2 - 2*r**3 + 2*(1 - r + r**2)**1.5
        ) / (
            np.sqrt(r)*(1+np.sqrt(r))
        )
        bimodal = abs(mean2-mean1) > separation_factor * \
            (np.sqrt(var1) + np.sqrt(var2))
        return bimodal


def dna_bimodality(
        input_csv_file, output_csv_path, output_jpg_path,
        input_zip_file: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`HelParBimodality <dna.dna_bimodality.HelParBimodality>` class and
    execute the :meth:`launch() <dna.dna_bimodality.HelParBimodality.launch>` method."""
    return HelParBimodality(**dict(locals())).launch()


dna_bimodality.__doc__ = HelParBimodality.__doc__
main = HelParBimodality.get_main(dna_bimodality, "Determine binormality/bimodality from a helical parameter dataset.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the CanalUnzip class and the command line interface."""
import re
import zipfile
import shutil
from typing import Optional

from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class CanalUnzip(BiobbObject):
    """
    | biobb_dna CanalUnzip
    | Tool for extracting biobb_canal output files.
    | Unzips a Canal output file contained within a zip file.

    Args:
        input_zip_file (str): Zip file with Canal output files. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/canal_output.zip>`_. Accepted formats: zip (edam:format_3987).
        output_path (str): Canal output file contained within input_zip_file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/canal_unzip_output.ser>`_. Accepted formats: ser (edam:format_2330), his (edam:format_3905), cor (edam:format_3465).
        output_list_path (str) (Optional): Text file with a list of all Canal output files contained within input_zip_file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/canal_unzip_output.txt>`_. Accepted formats: txt (edam:format_2330).
        properties (dic):
            * **type** (*str*) - (None) Type of file. Values: series, histo, corr.
            * **helpar_name** (*str*) - (None) Helical parameter name, only for 'series' and 'histo' types. Values: alphaC, alphaW, ampC, ampW, ax-bend, betaC, betaW, buckle, chiC, chiW, curv, deltaC, deltaW, epsilC, epsilW, gammaC, gammaW, h-ris, h-twi, inclin, majd, majw, mind, minw, opening, phaseC, phaseW, propel, reg, rise, roll, shear, shift, slide, stagger, stretch, tbend, tilt, tip, twist, xdisp, ydisp, zetaC, zetaW.
            * **correlation** (*str*) - (None) Correlation indexes separated by underscore (ie '98_165'), only for 'corr' type.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.curvesplus.canal_unzip import canal_unzip
            prop = {
                'type': 'series',
                'helpar_name': 'alphaC'
            }
            canal_unzip(
                input_zip_file='/path/to/canal/output.zip',
                output_path='/path/to/output.ser',
                output_list_path='/path/to/output.txt'
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_zip_file,
                 output_path, output_list_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_zip_file': input_zip_file
            },
            'out': {
                'output_path': output_path,
                'output_list_path': output_list_path
            }
        }

        # Properties specific for BB
        self.type = properties.get('type', None)
        self.helpar_name = properties.get('helpar_name', None)
        self.correlation = properties.get('correlation', None)
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`CanalUnzip <biobb_dna.curvesplus.canal_unzip.CanalUnzip>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # Check that both properties are set
        if self.type is None:
            fu.log("Property 'type' is mandatory to run CanalUnzip. Please set it.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that helpar_name is set if type is series or histo
        if self.type in ["series", "histo"] and self.helpar_name is None:
            fu.log("Property 'helpar_name' is mandatory to run CanalUnzip with type 'series' or 'histo'. Please set it.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that correlation is set if type is corr
        if self.type == "corr" and self.correlation is None:
            fu.log("Property 'correlation' is mandatory to run CanalUnzip with type 'corr'. Please set it.",
                   self.out_log, self.global_log)
            exit(1)

        extensions = {
            "series": "ser",
            "histo": "his",
            "corr": "cor"
        }
        # Check that the type is valid
        if self.type not in extensions:
            fu.log(f"Type {self.type} not valid. Valid types are: {', '.join(extensions.keys())}.",
                   self.out_log, self.global_log)
            exit(1)

        # generate sufix
        sufix = ""
        if self.type == "corr":
            # Check that the correlation is valid
            pattern = r'\d+_\d+'
            if not re.match(pattern, self.correlation):
                fu.log(f"Correlation {self.correlation} not valid. It should match the pattern <number_number>.",
                       self.out_log, self.global_log)
                exit(1)
            sufix = self.correlation
        else:
            # Check that the helpar_name is valid
            if self.helpar_name not in ["alphaC", "alphaW", "ampC", "ampW", "ax-bend", "betaC", "betaW", "buckle",
                                        "chiC", "chiW", "curv", "deltaC", "deltaW", "epsilC", "epsilW", "gammaC",
                                        "gammaW", "h-ris", "h-twi", "inclin", "majd", "majw", "mind", "minw",
                                        "opening", "phaseC", "phaseW", "propel", "reg", "rise", "roll", "shear",
                                        "shift", "slide", "stagger", "stretch", "tbend", "tilt", "tip", "twist",
                                        "xdisp", "ydisp", "zetaC", "zetaW"]:
                fu.log(f"Parameter {self.helpar_name} not valid. Valid parameters are: alphaC, alphaW, ampC, ampW, ax-bend, betaC, betaW, buckle, chiC, chiW, curv, deltaC, deltaW, epsilC, epsilW, gammaC, gammaW, h-ris, h-twi, inclin, majd, majw, mind, minw, opening, phaseC, phaseW, propel, reg, rise, roll, shear, shift, slide, stagger, stretch, tbend, tilt, tip, twist, xdisp, ydisp, zetaC, zetaW.",
                       self.out_log, self.global_log)
                exit(1)
            sufix = self.helpar_name

        # Generate the filename
        filename = f"canal_output_{sufix}.{extensions[self.type]}"

        # Unzip the file
        with zipfile.ZipFile(self.stage_io_dict["in"]["input_zip_file"], 'r') as zip_ref:
            # Check if the file exists in the zip file
            if filename in zip_ref.namelist():
                # Extract the file
                fu.log(f'{filename} exists, copying into {self.stage_io_dict["out"]["output_path"]}.',
                       self.out_log, self.global_log)
                with zip_ref.open(filename) as source, open(self.stage_io_dict["out"]["output_path"], "wb") as target:
                    shutil.copyfileobj(source, target)
            else:
                fu.log(f"File {filename} not found in the zip file.", self.out_log, self.global_log)
                exit(1)

            # Write the list of files
            if self.stage_io_dict["out"]["output_list_path"]:
                with open(self.stage_io_dict["out"]["output_list_path"], "w") as f:
                    for name in zip_ref.namelist():
                        f.write(f"{name}\n")

        # Run Biobb block
        # self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def canal_unzip(
        input_zip_file: str,
        output_path: str,
        output_list_path: Optional[str] = None,
        properties: Optional[dict] = None,
        **kwargs) -> int:
    """Create :class:`CanalUnzip <biobb_dna.curvesplus.canal_unzip.CanalUnzip>` class and
    execute the :meth:`launch() <biobb_dna.curvesplus.canal_unzip.CanalUnzip.launch>` method."""
    return CanalUnzip(**dict(locals())).launch()


canal_unzip.__doc__ = CanalUnzip.__doc__
main = CanalUnzip.get_main(canal_unzip, "Tool for extracting biobb_canal output files.")

if __name__ == '__main__':
    main()



from . import (
    backbone,
    curvesplus,
    dna,
    interbp_correlations,
    intrabp_correlations,
    stiffness,
)

name = "biobb_dna"
__all__ = [
    "dna",
    "curvesplus",
    "backbone",
    "stiffness",
    "interbp_correlations",
    "intrabp_correlations",
]
__version__ = "5.2.1"



#!/usr/bin/env python3

"""Module containing the DnaTimeseriesUnzip class and the command line interface."""
import re
import zipfile
import shutil
from typing import Optional

from biobb_dna.utils import constants
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class DnaTimeseriesUnzip(BiobbObject):
    """
    | biobb_dna DnaTimeseriesUnzip
    | Tool for extracting dna_timeseries output files.
    | Unzips a zip file containing dna_timeseries output files and extracts the csv and jpg files.

    Args:
        input_zip_file (str): Zip file with dna_timeseries output files. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/dna/timeseries_output.zip>`_. Accepted formats: zip (edam:format_3987).
        output_path_csv (str): dna_timeseries output csv file contained within input_zip_file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/dna_timeseries_unzip.csv>`_. Accepted formats: csv (edam:format_3752).
        output_path_jpg (str): dna_timeseries output jpg file contained within input_zip_file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/dna_timeseries_unzip.jpg>`_. Accepted formats: jpg (edam:format_3579).
        output_list_path (str) (Optional): Text file with a list of all dna_timeseries output files contained within input_zip_file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/dna_timeseries_unzip.txt>`_. Accepted formats: txt (edam:format_2330).
        properties (dic):
            * **type** (*str*) - (None) Type of analysis, series or histogram. Values: series, hist.
            * **parameter** (*str*) - (None) Type of parameter. Values: majd, majw, mind, minw, inclin, tip, xdisp, ydisp, shear, stretch, stagger, buckle, propel, opening, rise, roll, twist, shift, slide, tilt, alphaC, alphaW, betaC, betaW, gammaC, gammaW, deltaC, deltaW, epsilC, epsilW, zetaC, zetaW, chiC, chiW, phaseC, phaseW.
            * **sequence** (*str*) - (None) Nucleic acid sequence used for generating dna_timeseries output file.
            * **index** (*int*) - (1) Base pair index in the parameter 'sequence', starting from 1.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.dna.dna_timeseries_unzip import dna_timeseries_unzip
            prop = {
                'type': 'hist',
                'parameter': 'shift',
                'sequence': 'CGCGAATTCGCG',
                'index': 5
            }
            dna_timeseries_unzip(
                input_zip_file='/path/to/dna_timeseries/output.zip',
                output_path='/path/to/output.csv',
                output_list_path='/path/to/output.txt'
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_zip_file,
                 output_path_csv, output_path_jpg, output_list_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_zip_file': input_zip_file
            },
            'out': {
                'output_path_csv': output_path_csv,
                'output_path_jpg': output_path_jpg,
                'output_list_path': output_list_path
            }
        }

        # Properties specific for BB
        self.type = properties.get('type', None)
        self.parameter = properties.get('parameter', None)
        self.sequence = properties.get('sequence', None)
        self.index = properties.get('index', 1)
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`DnaTimeseriesUnzip <biobb_dna.dna.dna_timeseries_unzip.DnaTimeseriesUnzip>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # Check that both properties are set
        if self.type is None or self.parameter is None or self.sequence is None:
            fu.log("Properties 'type', 'parameter' and 'sequence' are mandatory to run DnaTimeseriesUnzip. Please set them.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that the type is valid
        if self.type not in ["series", "hist"]:
            fu.log(f"Type {self.type} not valid. Valid types are: series, hist.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that the parameter is valid
        if self.parameter not in constants.helical_parameters:
            fu.log(f"Parameter {self.parameter} not valid. Valid parameters are: {constants.helical_parameters}.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that the sequence is valid
        pattern = r'^[ACGT]+$'
        if not re.match(pattern, self.sequence):
            fu.log(f"Sequence {self.sequence} not valid. Only 'A', 'C', 'G' or 'T' bases allowed.",
                   self.out_log, self.global_log)
            exit(1)

        # Check that the index is valid
        if self.index < 1 or self.index >= len(self.sequence) - 1:
            fu.log(f"Index {self.index} not valid. It should be between 0 and {len(self.sequence) - 2}.",
                   self.out_log, self.global_log)
            exit(1)

        # Get index sequence base and next base
        bp = self.sequence[self.index-1] + self.sequence[self.index]

        # Get the filename
        filename = f"{self.type}_{self.parameter}_{self.index}_{bp}"
        csv_file = f"{filename}.csv"
        jpg_file = f"{filename}.jpg"

        # Unzip the file
        with zipfile.ZipFile(self.stage_io_dict["in"]["input_zip_file"], 'r') as zip_ref:
            # Check if the csv file exists in the zip file
            if csv_file in zip_ref.namelist():
                # Extract the file
                fu.log(f'{csv_file} exists, copying into {self.stage_io_dict["out"]["output_path_csv"]}.',
                       self.out_log, self.global_log)
                with zip_ref.open(csv_file) as source, open(self.stage_io_dict["out"]["output_path_csv"], "wb") as target:
                    shutil.copyfileobj(source, target)
            else:
                fu.log(f"File {csv_file} not found in the zip file.", self.out_log, self.global_log)
                exit(1)

            # Check if the jpg file exists in the zip file
            if jpg_file in zip_ref.namelist():
                # Extract the file
                fu.log(f'{jpg_file} exists, copying into {self.stage_io_dict["out"]["output_path_jpg"]}.',
                       self.out_log, self.global_log)
                with zip_ref.open(jpg_file) as source, open(self.stage_io_dict["out"]["output_path_jpg"], "wb") as target:
                    shutil.copyfileobj(source, target)
            else:
                fu.log(f"File {jpg_file} not found in the zip file.", self.out_log, self.global_log)
                exit(1)

            # Write the list of files
            if self.stage_io_dict["out"]["output_list_path"]:
                with open(self.stage_io_dict["out"]["output_list_path"], "w") as f:
                    for name in zip_ref.namelist():
                        f.write(f"{name}\n")

        # Run Biobb block
        # self.run_biobb()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def dna_timeseries_unzip(
        input_zip_file: str,
        output_path_csv: str,
        output_path_jpg: str,
        output_list_path: Optional[str] = None,
        properties: Optional[dict] = None,
        **kwargs) -> int:
    """Create :class:`DnaTimeseriesUnzip <biobb_dna.dna.dna_timeseries_unzip.DnaTimeseriesUnzip>` class and
    execute the :meth:`launch() <biobb_dna.dna.dna_timeseries_unzip.DnaTimeseriesUnzip.launch>` method."""
    return DnaTimeseriesUnzip(**dict(locals())).launch()


dna_timeseries_unzip.__doc__ = DnaTimeseriesUnzip.__doc__
main = DnaTimeseriesUnzip.get_main(dna_timeseries_unzip, "Tool for extracting dna_timeseries output files.")

if __name__ == '__main__':
    main()



from typing import Optional, Union


# TODO: Move this function to biobb_common.tools.file_utils
def _from_string_to_list(input_data: Optional[Union[str, list[str]]]) -> list[str]:
    """
    Converts a string to a list, splitting by commas or spaces. If the input is already a list, returns it as is.
    Returns an empty list if input_data is None.

    Parameters:
        input_data (str, list, or None): The string, list, or None value to convert.

    Returns:
        list: A list of string elements or an empty list if input_data is None.
    """
    if input_data is None:
        return []

    if isinstance(input_data, list):
        # If input is already a list, return it
        return input_data

    # If input is a string, determine the delimiter based on presence of commas
    delimiter = "," if "," in input_data else " "
    items = input_data.split(delimiter)

    # Remove whitespace from each item and ignore empty strings
    processed_items = [item.strip() for item in items if item.strip()]

    return processed_items



#!/usr/bin/env python3

"""Module containing the Canal class and the command line interface."""
import os
import zipfile
from typing import Optional
from pathlib import Path

from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class Canal(BiobbObject):
    """
    | biobb_dna Canal
    | Wrapper for the Canal executable that is part of the Curves+ software suite.
    | The Canal program is used to analyze the curvature of DNA structures.

    Args:
        input_cda_file (str): Input cda file, from Cur+ output. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/curves_output.cda>`_. Accepted formats: cda (edam:format_2330).
        input_lis_file (str) (Optional): Input lis file, from Cur+ output. File type: input. Accepted formats: lis (edam:format_2330).
        output_zip_path (str): zip filename for output files. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/canal_output.zip>`_. Accepted formats: zip (edam:format_3987).
        properties (dic):
            * **bases** (*str*) - (None) sequence of bases to be searched for in the I/P data (default is blank, meaning no specified sequence).
            * **itst** (*int*) - (0) Iteration start index.
            * **itnd** (*int*) - (0) Iteration end index.
            * **itdel** (*int*) - (1) Iteration delimiter.
            * **lev1** (*int*) - (0) Lower base level limit (i.e. base pairs) used for analysis.
            * **lev2** (*int*) - (0) Upper base level limit used for analysis. If lev1 > 0 and lev2 = 0, lev2 is set to lev1 (i.e. analyze lev1 only). If lev1=lev2=0, lev1 is set to 1 and lev2 is set to the length of the oligmer (i.e. analyze all levels).
            * **nastr** (*str*) - ('NA') character string used to indicate missing data in .ser files.
            * **cormin** (*float*) - (0.6) minimal absolute value for printing linear correlation coefficients between pairs of analyzed variables.
            * **series** (*bool*) - (False) if True then output spatial or time series data. Only possible for the analysis of single structures or single trajectories.
            * **histo** (*bool*) - (False) if True then output histogram data.
            * **corr** (*bool*) - (False) if True than output linear correlation coefficients between all variables.
            * **sequence** (*str*) - (Optional) sequence of the first strand of the corresponding DNA fragment, for each .cda file. If not given it will be parsed from .lis file.
            * **binary_path** (*str*) - ('Canal') Path to Canal executable, otherwise the program wil look for Canal executable in the binaries folder.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.curvesplus.biobb_canal import biobb_canal
            prop = {
                'series': 'True',
                'histo': 'True',
                'sequence': 'CGCGAATTCGCG'
            }
            biobb_canal(
                input_cda_file='/path/to/curves/output.cda',
                output_zip_path='/path/to/output.zip',
                properties=prop)
    Info:
        * wrapped_software:
            * name: Canal
            * version: >=2.6
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(self, input_cda_file, input_lis_file=None,
                 output_zip_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_cda_file': input_cda_file,
                'input_lis_file': input_lis_file,
            },
            'out': {
                'output_zip_path': output_zip_path
            }
        }

        # Properties specific for BB
        self.bases = properties.get('bases', None)
        self.nastr = properties.get('nastr', None)
        self.cormin = properties.get('cormin', 0.6)
        self.lev1 = properties.get('lev1', 0)
        self.lev2 = properties.get('lev2', 0)
        self.itst = properties.get('itst', 0)
        self.itnd = properties.get('itnd', 0)
        self.itdel = properties.get('itdel', 1)
        self.series = ".t." if properties.get('series', False) else ".f."
        self.histo = ".t." if properties.get('histo', False) else ".f."
        self.corr = ".t." if properties.get('corr', False) else ".f."
        self.sequence = properties.get('sequence', None)
        self.binary_path = properties.get('binary_path', 'Canal')
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`Canal <biobb_dna.curvesplus.biobb_canal.Canal>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        if self.sequence is None:
            if self.stage_io_dict['in']['input_lis_file'] is None:
                raise RuntimeError(
                    "if no sequence is passed in the configuration, "
                    "you must at least specify `input_lis_file` "
                    "so sequence can be parsed from there")
            lis_lines = Path(
                self.stage_io_dict['in']['input_lis_file']).read_text().splitlines()
            for line in lis_lines:
                if line.strip().startswith("Strand  1"):
                    self.sequence = line.split(" ")[-1]
                    fu.log(
                        f"using sequence {self.sequence} "
                        f"from {self.stage_io_dict['in']['input_lis_file']}",
                        self.out_log)

        # define temporary file name
        tmp_cda_path = Path(self.stage_io_dict['in']['input_cda_file']).name

        # change directory to temporary folder
        original_directory = os.getcwd()
        os.chdir(self.stage_io_dict.get("unique_dir", ""))

        # create intructions
        instructions = [
            f"{self.binary_path} <<! ",
            "&inp",
            "  lis=canal_output,"]
        if self.bases is not None:
            # add topology file if needed
            fu.log('Appending sequence of bases to be searched to command',
                   self.out_log, self.global_log)
            instructions.append(f"  seq={self.bases},")
        if self.nastr is not None:
            # add topology file if needed
            fu.log('Adding null values string specification to command',
                   self.out_log, self.global_log)
            instructions.append(f"  nastr={self.nastr},")

        instructions = instructions + [
            f"  cormin={self.cormin},",
            f"  lev1={self.lev1},lev2={self.lev2},",
            f"  itst={self.itst},itnd={self.itnd},itdel={self.itdel},",
            f"  histo={self.histo},",
            f"  series={self.series},",
            f"  corr={self.corr},",
            "&end",
            f"{tmp_cda_path} {self.sequence}",
            "!"]

        self.cmd = ["\n".join(instructions)]
        fu.log('Creating command line with instructions and required arguments',
               self.out_log, self.global_log)

        # Run Biobb block
        self.run_biobb()

        # change back to original directory
        os.chdir(original_directory)

        # create zipfile and write output inside
        zf = zipfile.ZipFile(
            Path(self.stage_io_dict["out"]["output_zip_path"]), "w")
        for canal_outfile in Path(self.stage_io_dict.get("unique_dir", "")).glob("canal_output*"):
            if canal_outfile.suffix not in (".zip"):
                zf.write(
                    canal_outfile,
                    arcname=canal_outfile.name)
        zf.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def biobb_canal(
        input_cda_file: str,
        output_zip_path: str,
        input_lis_file: Optional[str] = None,
        properties: Optional[dict] = None,
        **kwargs) -> int:
    """Create :class:`Canal <biobb_dna.curvesplus.biobb_canal.Canal>` class and
    execute the :meth:`launch() <biobb_dna.curvesplus.biobb_canal.Canal.launch>` method."""
    return Canal(**dict(locals())).launch()


biobb_canal.__doc__ = Canal.__doc__
main = Canal.get_main(biobb_canal, "Execute Canal from the Curves+ software suite.")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the HelParAverages class and the command line interface."""

from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger

from biobb_dna.utils import constants
from biobb_dna.utils.common import _from_string_to_list
from biobb_dna.utils.loader import read_series


class HelParAverages(BiobbObject):
    """
    | biobb_dna HelParAverages
    | Load .ser file for a given helical parameter and read each column corresponding to a base calculating average over each one.
    | Calculate average values for each base pair and save them in a .csv file.

    Args:
        input_ser_path (str): Path to .ser file for helical parameter. File is expected to be a table, with the first column being an index and the rest the helical parameter values for each base/basepair. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/dna/canal_output_shift.ser>`_. Accepted formats: ser (edam:format_2330).
        output_csv_path (str): Path to .csv file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/shift_avg.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/shift_avg.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **sequence** (*str*) - (None) Nucleic acid sequence corresponding to the input .ser file. Length of sequence is expected to be the same as the total number of columns in the .ser file, minus the index column (even if later on a subset of columns is selected with the *seqpos* option).
            * **helpar_name** (*str*) - (Optional) helical parameter name.
            * **stride** (*int*) - (1000) granularity of the number of snapshots for plotting time series.
            * **seqpos** (*list*) - (None) list of sequence positions (columns indices starting by 0) to analyze.  If not specified it will analyse the complete sequence.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.dna.dna_averages import dna_averages

            prop = {
                'helpar_name': 'twist',
                'seqpos': [1,2],
                'sequence': 'GCAT'
            }
            dna_averages(
                input_ser_path='/path/to/twist.ser',
                output_csv_path='/path/to/table/output.csv',
                output_jpg_path='/path/to/table/output.jpg',
                properties=prop)

    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
        self,
        input_ser_path,
        output_csv_path,
        output_jpg_path,
        properties=None,
        **kwargs,
    ) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_ser_path": input_ser_path,
            },
            "out": {
                "output_csv_path": output_csv_path,
                "output_jpg_path": output_jpg_path,
            },
        }

        # Properties specific for BB
        self.properties = properties
        self.sequence = properties.get("sequence", None)
        self.stride = properties.get("stride", 1000)
        self.seqpos = [
            int(elem) for elem in _from_string_to_list(properties.get("seqpos", None))
        ]
        self.helpar_name = properties.get("helpar_name", None)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`HelParAverages <dna.averages.HelParAverages>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # check sequence
        if self.sequence is None or len(self.sequence) < 2:
            raise ValueError("sequence is null or too short!")

        # get helical parameter from filename if not specified
        if self.helpar_name is None:
            for hp in constants.helical_parameters:
                ser_name = Path(self.stage_io_dict["in"]["input_ser_path"]).name.lower()
                if hp.lower() in ser_name:
                    self.helpar_name = hp
            if self.helpar_name is None:
                raise ValueError(
                    "Helical parameter name can't be inferred from file, "
                    "so it must be specified!"
                )
        else:
            if self.helpar_name not in constants.helical_parameters:
                raise ValueError(
                    "Helical parameter name is invalid! "
                    f"Options: {constants.helical_parameters}"
                )

        # get base length and unit from helical parameter name
        if self.helpar_name.lower() in constants.hp_basepairs:
            self.baselen = 1
        elif self.helpar_name.lower() in constants.hp_singlebases:
            self.baselen = 0
        if self.helpar_name in constants.hp_angular:
            self.hp_unit = "Degrees"
        else:
            self.hp_unit = "Angstroms"

        # check seqpos
        if self.seqpos:
            if (max(self.seqpos) > len(self.sequence) - 2) or (min(self.seqpos) < 1):
                raise ValueError(
                    f"seqpos values must be between 1 and {len(self.sequence) - 2}"
                )
            if not (isinstance(self.seqpos, list) and len(self.seqpos) > 1):
                raise ValueError("seqpos must be a list of at least two integers")
        else:
            self.seqpos = None  # type: ignore

        # read input .ser file
        ser_data = read_series(
            self.stage_io_dict["in"]["input_ser_path"], usecols=self.seqpos
        )
        if not self.seqpos:
            ser_data = ser_data[ser_data.columns[1:-1]]
            # discard first and last base(pairs) from sequence
            sequence = self.sequence[1:]
            xlabels = [
                f"{sequence[i:i+1+self.baselen]}"
                for i in range(len(ser_data.columns) - self.baselen)
            ]
        else:
            sequence = self.sequence
            xlabels = [f"{sequence[i:i+1+self.baselen]}" for i in self.seqpos]

        # rename duplicated subunits
        while any(pd.Index(ser_data.columns).duplicated()):
            ser_data.columns = [
                name if not duplicated else name + "_dup"
                for duplicated, name in zip(
                    pd.Index(ser_data.columns).duplicated(), ser_data.columns
                )
            ]

        # write output files for all selected bases
        means = ser_data.mean(axis=0).iloc[: len(xlabels)]
        stds = ser_data.std(axis=0).iloc[: len(xlabels)]

        # save plot
        fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
        axs.errorbar(
            means.index, means.to_numpy(), yerr=stds.to_numpy(), marker="o", capsize=5
        )
        axs.set_xticks(means.index)
        axs.set_xticklabels(xlabels, rotation=90)
        axs.set_xlabel("Sequence Base Pair " f"{'Step' if self.baselen == 1 else ''}")
        axs.set_ylabel(f"{self.helpar_name.capitalize()} ({self.hp_unit})")
        axs.set_title(
            "Base Pair "
            f"{'Step' if self.baselen == 1 else ''} "
            f"Helical Parameter: {self.helpar_name.capitalize()}"
        )
        fig.savefig(self.stage_io_dict["out"]["output_jpg_path"], format="jpg")

        # save table
        dataset = pd.DataFrame(
            {
                f"Base Pair {'Step' if self.baselen == 1 else ''}": xlabels,
                "mean": means.to_numpy(),
                "std": stds.to_numpy(),
            }
        )
        dataset.to_csv(self.stage_io_dict["out"]["output_csv_path"], index=False)

        plt.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0


def dna_averages(
    input_ser_path: str,
    output_csv_path: str,
    output_jpg_path: str,
    properties: Optional[dict] = None,
    **kwargs,
) -> int:
    """Create :class:`HelParAverages <dna.dna_averages.HelParAverages>` class and
    execute the :meth:`launch() <dna.dna_averages.HelParAverages.launch>` method."""
    return HelParAverages(**dict(locals())).launch()


dna_averages.__doc__ = HelParAverages.__doc__
main = HelParAverages.get_main(dna_averages, "Load helical parameter file and calculate average values for each base pair.")

if __name__ == '__main__':
    main()




hp_singlebases = [
    "shear",
    "stagger",
    "stretch",
    "buckle",
    "opening",
    "propel",
    "inclin",
    "tip",
    "xdisp",
    "ydisp"
]

hp_basepairs = [
    "majw",
    "majd",
    "minw",
    "mind",
    "rise",
    "shift",
    "slide",
    "roll",
    "tilt",
    "twist"
]

hp_angular = [
    "roll",
    "tilt",
    "twist",
    "buckle",
    "opening",
    "propel",
    "inclin",
    "tip"
]

hp_translational = [
    "shear",
    "stagger",
    "stretch",
    "xdisp",
    "ydisp",
    "majw",
    "majd",
    "minw",
    "mind",
    "rise",
    "shift",
    "slide"
]

hp_backbone = [
    "alphaC",
    "alphaW",
    "betaC",
    "betaW",
    "gammaC",
    "gammaW",
    "deltaC",
    "deltaW",
    "chiC",
    "chiW",
    "phaseC",
    "phaseW",
    "epsilC",
    "epsilW",
    "zetaC",
    "zetaW"
]

helical_parameters = hp_basepairs + hp_singlebases + hp_backbone



#!/usr/bin/env python3

"""Module containing the InterHelParCorrelation class and the command line interface."""
from typing import Optional

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger
from biobb_dna.utils.loader import load_data


class InterHelParCorrelation(BiobbObject):
    """
    | biobb_dna InterHelParCorrelation
    | Calculate correlation between helical parameters for a single inter-base pair.
    | Calculate correlation between helical parameters for a single inter-base pair.

    Args:
        input_filename_shift (str): Path to .csv file with data for helical parameter 'shift'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_shift_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_slide (str): Path to .csv file with data for helical parameter 'slide'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_slide_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_rise (str): Path to .csv file with data for helical parameter 'rise'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_rise_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_tilt (str): Path to .csv file with data for helical parameter 'tilt'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_tilt_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_roll (str): Path to .csv file with data for helical parameter 'roll'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_roll_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        input_filename_twist (str): Path to .csv file with data for helical parameter 'twist'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/stiffness/series_twist_AA.csv>`_. Accepted formats: csv (edam:format_3752).
        output_csv_path (str): Path to directory where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/inter_hpcorr_ref.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/inter_hpcorr_ref.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **basepair** (*str*) - (None) Name of basepair analyzed.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.interbp_correlations.interhpcorr import interhpcorr

            prop = {
                'basepair': 'AA',
            }
            interhpcorr(
                input_filename_shift='path/to/shift.csv',
                input_filename_slide='path/to/slide.csv',
                input_filename_rise='path/to/rise.csv',
                input_filename_tilt='path/to/tilt.csv',
                input_filename_roll='path/to/roll.csv',
                input_filename_twist='path/to/twist.csv',
                output_csv_path='path/to/output/file.csv',
                output_jpg_path='path/to/output/file.jpg',
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
            self, input_filename_shift, input_filename_slide,
            input_filename_rise, input_filename_tilt,
            input_filename_roll, input_filename_twist,
            output_csv_path, output_jpg_path,
            properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_filename_shift': input_filename_shift,
                'input_filename_slide': input_filename_slide,
                'input_filename_rise': input_filename_rise,
                'input_filename_tilt': input_filename_tilt,
                'input_filename_roll': input_filename_roll,
                'input_filename_twist': input_filename_twist
            },
            'out': {
                'output_csv_path': output_csv_path,
                'output_jpg_path': output_jpg_path
            }
        }

        self.properties = properties
        self.basepair = properties.get("basepair", None)

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`InterHelParCorrelation <interbp_correlations.interhpcorr.InterHelParCorrelation>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # read input
        shift = load_data(self.stage_io_dict["in"]["input_filename_shift"])
        slide = load_data(self.stage_io_dict["in"]["input_filename_slide"])
        rise = load_data(self.stage_io_dict["in"]["input_filename_rise"])
        tilt = load_data(self.stage_io_dict["in"]["input_filename_tilt"])
        roll = load_data(self.stage_io_dict["in"]["input_filename_roll"])
        twist = load_data(self.stage_io_dict["in"]["input_filename_twist"])

        # get basepair
        if self.basepair is None:
            self.basepair = shift.columns[0]

        # make matrix
        coordinates = ["shift", "slide", "rise", "tilt", "roll", "twist"]
        corr_matrix = pd.DataFrame(
            np.eye(6, 6), index=coordinates, columns=coordinates)

        # shift
        # corr_matrix["shift"]["slide"] = shift.corrwith(slide, method="pearson")
        corr_matrix.loc["slide", "shift"] = shift.corrwith(slide, method="pearson").values[0]
        # corr_matrix["shift"]["rise"] = shift.corrwith(rise, method="pearson")
        corr_matrix.loc["rise", "shift"] = shift.corrwith(rise, method="pearson").values[0]
        # corr_matrix["shift"]["tilt"] = shift.corrwith(tilt, method=self.circlineal)
        corr_matrix.loc["tilt", "shift"] = shift.corrwith(tilt, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["shift"]["roll"] = shift.corrwith(roll, method=self.circlineal)
        corr_matrix.loc["roll", "shift"] = shift.corrwith(roll, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["shift"]["twist"] = shift.corrwith(twist, method=self.circlineal)
        corr_matrix.loc["twist", "shift"] = shift.corrwith(twist, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["slide"]["shift"] = corr_matrix["shift"]["slide"]
        corr_matrix.loc["shift", "slide"] = corr_matrix.loc["slide", "shift"]
        # corr_matrix["rise"]["shift"] = corr_matrix["shift"]["rise"]
        corr_matrix.loc["shift", "rise"] = corr_matrix.loc["rise", "shift"]
        # corr_matrix["tilt"]["shift"] = corr_matrix["shift"]["tilt"]
        corr_matrix.loc["shift", "tilt"] = corr_matrix.loc["tilt", "shift"]
        # corr_matrix["roll"]["shift"] = corr_matrix["shift"]["roll"]
        corr_matrix.loc["shift", "roll"] = corr_matrix.loc["roll", "shift"]
        # corr_matrix["twist"]["shift"] = corr_matrix["shift"]["twist"]
        corr_matrix.loc["shift", "twist"] = corr_matrix.loc["twist", "shift"]

        # slide
        # corr_matrix["slide"]["rise"] = slide.corrwith(rise, method="pearson")
        corr_matrix.loc["rise", "slide"] = slide.corrwith(rise, method="pearson").values[0]
        # corr_matrix["slide"]["tilt"] = slide.corrwith(tilt, method=self.circlineal)
        corr_matrix.loc["tilt", "slide"] = slide.corrwith(tilt, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["slide"]["roll"] = slide.corrwith(roll, method=self.circlineal)
        corr_matrix.loc["roll", "slide"] = slide.corrwith(roll, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["slide"]["twist"] = slide.corrwith(twist, method=self.circlineal)
        corr_matrix.loc["twist", "slide"] = slide.corrwith(twist, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["rise"]["slide"] = corr_matrix["slide"]["rise"]
        corr_matrix.loc["slide", "rise"] = corr_matrix.loc["rise", "slide"]
        # corr_matrix["tilt"]["slide"] = corr_matrix["slide"]["tilt"]
        corr_matrix.loc["slide", "tilt"] = corr_matrix.loc["tilt", "slide"]
        # corr_matrix["roll"]["slide"] = corr_matrix["slide"]["roll"]
        corr_matrix.loc["slide", "roll"] = corr_matrix.loc["roll", "slide"]
        # corr_matrix["twist"]["slide"] = corr_matrix["slide"]["twist"]
        corr_matrix.loc["slide", "twist"] = corr_matrix.loc["twist", "slide"]

        # rise
        # corr_matrix["rise"]["tilt"] = rise.corrwith(tilt, method=self.circlineal)
        corr_matrix.loc["tilt", "rise"] = rise.corrwith(tilt, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["rise"]["roll"] = rise.corrwith(roll, method=self.circlineal)
        corr_matrix.loc["roll", "rise"] = rise.corrwith(roll, method=self.circlineal).values[0]  # type: ignore
        # corr_matrix["rise"]["twist"] = rise.corrwith(twist, method=self.circlineal)
        corr_matrix.loc["twist", "rise"] = rise.corrwith(twist, method=self.circlineal).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["tilt"]["rise"] = corr_matrix["rise"]["tilt"]
        corr_matrix.loc["rise", "tilt"] = corr_matrix.loc["tilt", "rise"]
        # corr_matrix["roll"]["rise"] = corr_matrix["rise"]["roll"]
        corr_matrix.loc["rise", "roll"] = corr_matrix.loc["roll", "rise"]
        # corr_matrix["twist"]["rise"] = corr_matrix["rise"]["twist"]
        corr_matrix.loc["rise", "twist"] = corr_matrix.loc["twist", "rise"]

        # tilt
        # corr_matrix["tilt"]["roll"] = tilt.corrwith(roll, method=self.circular)
        corr_matrix.loc["roll", "tilt"] = tilt.corrwith(roll, method=self.circular).values[0]  # type: ignore
        # corr_matrix["tilt"]["twist"] = tilt.corrwith(twist, method=self.circular)
        corr_matrix.loc["twist", "tilt"] = tilt.corrwith(twist, method=self.circular).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["roll"]["tilt"] = corr_matrix["tilt"]["roll"]
        corr_matrix.loc["tilt", "roll"] = corr_matrix.loc["roll", "tilt"]
        # corr_matrix["twist"]["tilt"] = corr_matrix["tilt"]["twist"]
        corr_matrix.loc["tilt", "twist"] = corr_matrix.loc["twist", "tilt"]

        # roll
        # corr_matrix["roll"]["twist"] = roll.corrwith(twist, method=self.circular)
        corr_matrix.loc["twist", "roll"] = roll.corrwith(twist, method=self.circular).values[0]  # type: ignore
        # symmetric values
        # corr_matrix["twist"]["roll"] = corr_matrix["roll"]["twist"]
        corr_matrix.loc["roll", "twist"] = corr_matrix.loc["twist", "roll"]

        # save csv data
        corr_matrix.to_csv(self.stage_io_dict["out"]["output_csv_path"])

        # create heatmap
        fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
        axs.pcolor(corr_matrix)
        # Loop over data dimensions and create text annotations.
        for i in range(len(corr_matrix)):
            for j in range(len(corr_matrix)):
                axs.text(
                    j+.5,
                    i+.5,
                    f"{corr_matrix[coordinates[j]].loc[coordinates[i]]:.2f}",
                    ha="center",
                    va="center",
                    color="w")
        axs.set_xticks([i + 0.5 for i in range(len(corr_matrix))])
        axs.set_xticklabels(corr_matrix.columns, rotation=90)
        axs.set_yticks([i+0.5 for i in range(len(corr_matrix))])
        axs.set_yticklabels(corr_matrix.index)
        axs.set_title(
            "Helical Parameter Correlation "
            f"for Base Pair Step \'{self.basepair}\'")
        fig.tight_layout()
        fig.savefig(
            self.stage_io_dict['out']['output_jpg_path'],
            format="jpg")
        plt.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0

    def get_corr_method(self, corrtype1, corrtype2):
        if corrtype1 == "circular" and corrtype2 == "linear":
            method = self.circlineal
        if corrtype1 == "linear" and corrtype2 == "circular":
            method = self.circlineal
        elif corrtype1 == "circular" and corrtype2 == "circular":
            method = self.circular
        else:
            method = "pearson"
        return method

    @staticmethod
    def circular(x1, x2):
        x1 = x1 * np.pi / 180
        x2 = x2 * np.pi / 180
        diff_1 = np.sin(x1 - x1.mean())
        diff_2 = np.sin(x2 - x2.mean())
        num = (diff_1 * diff_2).sum()
        den = np.sqrt((diff_1 ** 2).sum() * (diff_2 ** 2).sum())
        return num / den

    @staticmethod
    def circlineal(x1, x2):
        x2 = x2 * np.pi / 180
        rc = np.corrcoef(x1, np.cos(x2))[1, 0]
        rs = np.corrcoef(x1, np.sin(x2))[1, 0]
        rcs = np.corrcoef(np.sin(x2), np.cos(x2))[1, 0]
        num = (rc ** 2) + (rs ** 2) - 2 * rc * rs * rcs
        den = 1 - (rcs ** 2)
        correlation = np.sqrt(num / den)
        if np.corrcoef(x1, x2)[1, 0] < 0:
            correlation *= -1
        return correlation


def interhpcorr(
        input_filename_shift: str, input_filename_slide: str,
        input_filename_rise: str, input_filename_tilt: str,
        input_filename_roll: str, input_filename_twist: str,
        output_csv_path: str, output_jpg_path: str,
        properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`InterHelParCorrelation <interbp_correlations.interhpcorr.InterHelParCorrelation>` class and
    execute the :meth:`launch() <interbp_correlations.interhpcorr.InterHelParCorrelation.launch>` method."""
    return InterHelParCorrelation(**dict(locals())).launch()


interhpcorr.__doc__ = InterHelParCorrelation.__doc__
main = InterHelParCorrelation.get_main(interhpcorr, "Load helical parameter file and save base data individually.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Utility functions to load files."""
import zipfile
from pathlib import Path

import pandas as pd


def read_series(input_serfile, usecols=None):
    """Read .ser file"""
    extra_kwargs = dict(
        header=None,
        sep='\\s+',
        index_col=0)
    ser_data = pd.read_csv(input_serfile, **extra_kwargs)  # type: ignore
    if usecols is not None:
        if 0 in usecols:
            usecols.pop(usecols.index(0))
        ser_data = ser_data[[i+1 for i in usecols]]
    return ser_data


def load_data(data_filename, inner_file=None):
    """Read .csv file directly or from inside a .zip file."""
    if Path(data_filename).suffix == ".zip":
        zf = zipfile.ZipFile(data_filename, "r")
        # use provided data filename of look for csv file
        if inner_file is not None:
            dataset = zf.open(inner_file)
        else:
            print(
                "inner file name not provided, "
                "using first .csv file found inside .zip.")
            for fn in zf.infolist():
                if fn.filename.endswith(".csv"):
                    dataset = zf.open(fn)
                    break
    elif Path(data_filename).suffix == ".csv":
        dataset = data_filename
    else:
        raise IOError("input file extension must be .zip or .csv!")
    data = pd.read_csv(dataset, index_col=0)
    return data



from . import dna_averages
from . import dna_bimodality
from . import dna_timeseries
from . import dna_timeseries_unzip
name = "dna"
__all__ = ["dna_averages", "dna_bimodality", "dna_timeseries", "dna_timeseries_unzip"]



#!/usr/bin/env python3
"""Module containing the CanonicalAG class and the command line interface."""

from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger

from biobb_dna.utils.common import _from_string_to_list
from biobb_dna.utils.loader import read_series
from biobb_dna.utils.transform import inverse_complement


class CanonicalAG(BiobbObject):
    """
    | biobb_dna CanonicalAG
    | Calculate Canonical Alpha/Gamma populations from alpha and gamma parameters.
    | Calculate Canonical Alpha/Gamma populations from alpha and gamma parameters.

    Args:
        input_alphaC_path (str): Path to .ser file for helical parameter 'alphaC'. File type: input. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/backbone/canal_output_alphaC.ser>`_. Accepted formats: ser (edam:format_2330).
        input_alphaW_path (str): Path to .ser file for helical parameter 'alphaW'. File type: input. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/backbone/canal_output_alphaW.ser>`_. Accepted formats: ser (edam:format_2330).
        input_gammaC_path (str): Path to .ser file for helical parameter 'gammaC'. File type: input. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/backbone/canal_output_gammaC.ser>`_. Accepted formats: ser (edam:format_2330).
        input_gammaW_path (str): Path to .ser file for helical parameter 'gammaW'. File type: input. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/backbone/canal_output_gammaW.ser>`_. Accepted formats: ser (edam:format_2330).
        output_csv_path (str): Path to .csv file where output is saved. File type: output. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/backbone/canonag_ref.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/backbone/canonag_ref.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **sequence** (*str*) - (None) Nucleic acid sequence corresponding to the input .ser file. Length of sequence is expected to be the same as the total number of columns in the .ser file, minus the index column (even if later on a subset of columns is selected with the *seqpos* option).
            * **seqpos** (*list*) - (None) list of sequence positions (columns indices starting by 0) to analyze.  If not specified it will analyse the complete sequence.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.backbone.canonicalag import canonicalag

            prop = {
                'seqpos': [1,2],
                'sequence': 'GCAT',
            }
            canonicalag(
                input_alphaC_path='/path/to/alphaC.ser',
                input_alphaW_path='/path/to/alphaW.ser',
                input_gammaC_path='/path/to/gammaC.ser',
                input_gammaW_path='/path/to/gammaW.ser',
                output_csv_path='/path/to/table/output.csv',
                output_jpg_path='/path/to/table/output.jpg',
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
        self,
        input_alphaC_path,
        input_alphaW_path,
        input_gammaC_path,
        input_gammaW_path,
        output_csv_path,
        output_jpg_path,
        properties=None,
        **kwargs,
    ) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_alphaC_path": input_alphaC_path,
                "input_alphaW_path": input_alphaW_path,
                "input_gammaC_path": input_gammaC_path,
                "input_gammaW_path": input_gammaW_path,
            },
            "out": {
                "output_csv_path": output_csv_path,
                "output_jpg_path": output_jpg_path,
            },
        }

        self.properties = properties
        self.sequence = properties.get("sequence")
        self.stride = properties.get("stride", 1000)
        self.seqpos = [
            int(elem) for elem in _from_string_to_list(properties.get("seqpos", None))
        ]

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`CanonicalAG <backbone.canonicalag.CanonicalAG>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # check sequence
        if self.sequence is None or len(self.sequence) < 2:
            raise ValueError("sequence is null or too short!")

        # check seqpos
        if self.seqpos:
            if (max(self.seqpos) > len(self.sequence) - 2) or (min(self.seqpos) < 1):
                raise ValueError(
                    f"seqpos values must be between 1 and {len(self.sequence) - 2}"
                )
            if not (isinstance(self.seqpos, list) and len(self.seqpos) > 1):
                raise ValueError("seqpos must be a list of at least two integers")
        else:
            self.seqpos = None  # type: ignore

        # read input files
        alphaC = read_series(
            self.stage_io_dict["in"]["input_alphaC_path"], usecols=self.seqpos
        )
        alphaW = read_series(
            self.stage_io_dict["in"]["input_alphaW_path"], usecols=self.seqpos
        )
        gammaC = read_series(
            self.stage_io_dict["in"]["input_gammaC_path"], usecols=self.seqpos
        )
        gammaW = read_series(
            self.stage_io_dict["in"]["input_gammaW_path"], usecols=self.seqpos
        )

        # fix angle range so its not negative
        alphaC = self.fix_angles(alphaC)
        alphaW = self.fix_angles(alphaW)
        gammaC = self.fix_angles(gammaC)
        gammaW = self.fix_angles(gammaW)

        # calculate difference between epsil and zeta parameters
        xlabels = self.get_xlabels(self.sequence, inverse_complement(self.sequence))
        canonical_populations = self.check_alpha_gamma(alphaC, gammaC, alphaW, gammaW)

        # save table
        canonical_populations.name = "Canonical alpha/gamma"
        ag_populations_df = pd.DataFrame(
            {"Nucleotide": xlabels, "Canonical alpha/gamma": canonical_populations}
        )
        ag_populations_df.to_csv(
            self.stage_io_dict["out"]["output_csv_path"], index=False
        )

        # save plot
        fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
        axs.bar(
            range(len(xlabels)), canonical_populations, label="canonical alpha/gamma"
        )
        axs.bar(
            range(len(xlabels)),
            100 - canonical_populations,
            bottom=canonical_populations,
            label=None,
        )
        # empty bar to divide both sequences
        axs.bar([len(alphaC.columns)], [100], color="white", label=None)
        axs.legend()
        axs.set_xticks(range(len(xlabels)))
        axs.set_xticklabels(xlabels, rotation=90)
        axs.set_xlabel("Nucleotide Sequence")
        axs.set_ylabel("Canonical Alpha-Gamma (%)")
        axs.set_title("Nucleotide parameter: Canonical Alpha-Gamma")
        fig.savefig(self.stage_io_dict["out"]["output_jpg_path"], format="jpg")
        plt.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        return 0

    def get_xlabels(self, strand1, strand2):
        # get list of tetramers, except first and last two bases
        labelsW = list(strand1)
        labelsW[0] = f"{labelsW[0]}5'"
        labelsW[-1] = f"{labelsW[-1]}3'"
        labelsW = [f"{i}-{j}" for i, j in zip(labelsW, range(1, len(labelsW) + 1))]
        labelsC = list(strand2)[::-1]
        labelsC[0] = f"{labelsC[0]}5'"
        labelsC[-1] = f"{labelsC[-1]}3'"
        labelsC = [f"{i}-{j}" for i, j in zip(labelsC, range(len(labelsC), 0, -1))]

        if self.seqpos:
            labelsC = [labelsC[i] for i in self.seqpos]
            labelsW = [labelsW[i] for i in self.seqpos]
        xlabels = labelsW + ["-"] + labelsC
        return xlabels

    def check_alpha_gamma(self, alphaC, gammaC, alphaW, gammaW):
        separator_df = pd.DataFrame({"-": np.nan}, index=range(len(gammaW)))
        gamma = pd.concat([gammaW, separator_df, gammaC[gammaC.columns[::-1]]], axis=1)
        alpha = pd.concat([alphaW, separator_df, alphaC[alphaC.columns[::-1]]], axis=1)

        alpha_filter = np.logical_and(alpha > 240, alpha < 360)
        gamma_filter = np.logical_and(gamma > 0, gamma < 120)
        canonical_alpha_gamma = np.logical_and(alpha_filter, gamma_filter).mean() * 100

        return canonical_alpha_gamma

    def fix_angles(self, dataset):
        values = np.where(dataset < 0, dataset + 360, dataset)
        values = np.where(values > 360, values - 360, values)
        dataset = pd.DataFrame(values)
        return dataset


def canonicalag(
    input_alphaC_path: str,
    input_alphaW_path: str,
    input_gammaC_path: str,
    input_gammaW_path: str,
    output_csv_path: str,
    output_jpg_path: str,
    properties: Optional[dict] = None,
    **kwargs,
) -> int:
    """Create :class:`CanonicalAG <dna.backbone.canonicalag.CanonicalAG>` class and
    execute the: meth: `launch() <dna.backbone.canonicalag.CanonicalAG.launch>` method."""
    return CanonicalAG(**dict(locals())).launch()


canonicalag.__doc__ = CanonicalAG.__doc__
main = CanonicalAG.get_main(canonicalag, "Calculate Canonical Alpha/Gamma distributions.")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Module containing the IntraBasePairCorrelation class and the command line interface."""
from itertools import product
from typing import Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools.file_utils import launchlogger

from biobb_dna.utils import constants
from biobb_dna.utils.common import _from_string_to_list
from biobb_dna.utils.loader import read_series


class IntraBasePairCorrelation(BiobbObject):
    """
    | biobb_dna IntraBasePairCorrelation
    | Calculate correlation between all intra-base pairs of a single sequence and for a single helical parameter.
    | Calculate correlation between neighboring base pairs and pairs of helical parameters.

    Args:
        input_filename_shear (str): Path to .ser file with data for helical parameter 'shear'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_shear.ser>`_. Accepted formats: ser (edam:format_2330).
        input_filename_stretch (str): Path to .ser file with data for helical parameter 'stretch'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_stretch.ser>`_. Accepted formats: ser (edam:format_2330).
        input_filename_stagger (str): Path to .ser file with data for helical parameter 'stagger'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_stagger.ser>`_. Accepted formats: ser (edam:format_2330).
        input_filename_buckle (str): Path to .ser file with data for helical parameter 'buckle'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_buckle.ser>`_. Accepted formats: ser (edam:format_2330).
        input_filename_propel (str): Path to .ser file with data for helical parameter 'propel'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_propel.ser>`_. Accepted formats: ser (edam:format_2330).
        input_filename_opening (str): Path to .ser file with data for helical parameter 'opening'. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/correlation/canal_output_opening.ser>`_. Accepted formats: ser (edam:format_2330).
        output_csv_path (str): Path to directory where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/intra_bpcorr_ref.csv>`_. Accepted formats: csv (edam:format_3752).
        output_jpg_path (str): Path to .jpg file where output is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/correlation/intra_bpcorr_ref.jpg>`_. Accepted formats: jpg (edam:format_3579).
        properties (dict):
            * **sequence** (*str*) - (None) Nucleic acid sequence for the input .ser file. Length of sequence is expected to be the same as the total number of columns in the .ser file, minus the index column (even if later on a subset of columns is selected with the *seqpos* option).
            * **seqpos** (*list*) - (None) list of sequence positions (columns indices starting by 0) to analyze.  If not specified it will analyse the complete sequence.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.intrabp_correlations.intrabpcorr import intrabpcorr

            intrabpcorr(
                input_filename_shear='path/to/input/shear.ser',
                input_filename_stretch='path/to/input/stretch.ser',
                input_filename_stagger='path/to/input/stagger.ser',
                input_filename_buckle='path/to/input/buckle.ser',
                input_filename_propel='path/to/input/propel.ser',
                input_filename_opening='path/to/input/opening.ser',
                output_csv_path='path/to/output/file.csv',
                output_jpg_path='path/to/output/plot.jpg',
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
        self,
        input_filename_shear,
        input_filename_stretch,
        input_filename_stagger,
        input_filename_buckle,
        input_filename_propel,
        input_filename_opening,
        output_csv_path,
        output_jpg_path,
        properties=None,
        **kwargs,
    ) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_filename_shear": input_filename_shear,
                "input_filename_stretch": input_filename_stretch,
                "input_filename_stagger": input_filename_stagger,
                "input_filename_buckle": input_filename_buckle,
                "input_filename_propel": input_filename_propel,
                "input_filename_opening": input_filename_opening,
            },
            "out": {
                "output_csv_path": output_csv_path,
                "output_jpg_path": output_jpg_path,
            },
        }

        self.properties = properties
        self.sequence = properties.get("sequence", None)
        self.seqpos = [
            int(elem) for elem in _from_string_to_list(properties.get("seqpos", None))
        ]

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`HelParCorrelation <intrabp_correlations.intrabpcorr.IntraBasePairCorrelation>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # check sequence
        if self.sequence is None or len(self.sequence) < 2:
            raise ValueError("sequence is null or too short!")

        # check seqpos
        if self.seqpos:
            if not (isinstance(self.seqpos, list) and len(self.seqpos) > 1):
                raise ValueError("seqpos must be a list of at least two integers")
        else:
            self.seqpos = None  # type: ignore

        # read input
        shear = read_series(
            self.stage_io_dict["in"]["input_filename_shear"], usecols=self.seqpos
        )
        stretch = read_series(
            self.stage_io_dict["in"]["input_filename_stretch"], usecols=self.seqpos
        )
        stagger = read_series(
            self.stage_io_dict["in"]["input_filename_stagger"], usecols=self.seqpos
        )
        buckle = read_series(
            self.stage_io_dict["in"]["input_filename_buckle"], usecols=self.seqpos
        )
        propel = read_series(
            self.stage_io_dict["in"]["input_filename_propel"], usecols=self.seqpos
        )
        opening = read_series(
            self.stage_io_dict["in"]["input_filename_opening"], usecols=self.seqpos
        )

        if not self.seqpos:
            # drop first and last columns
            shear = shear[shear.columns[1:-1]]
            stretch = stretch[stretch.columns[1:-1]]
            stagger = stagger[stagger.columns[1:-1]]
            buckle = buckle[buckle.columns[1:-1]]
            propel = propel[propel.columns[1:-1]]
            opening = opening[opening.columns[1:-1]]
            labels = [
                f"{i+1}_{self.sequence[i:i+1]}"
                for i in range(1, len(shear.columns) + 1)
            ]
            corr_index = [
                f"{self.sequence[i:i+2]}" for i in range(1, len(shear.columns) + 1)
            ]
        else:
            labels = [f"{i+1}_{self.sequence[i:i+1]}" for i in self.seqpos]
            corr_index = [f"{self.sequence[i:i+2]}" for i in self.seqpos]

        # rename duplicated subunits
        shear.columns = labels
        stretch.columns = labels
        stagger.columns = labels
        buckle.columns = labels
        propel.columns = labels
        opening.columns = labels

        # set names to each dataset
        shear.name = "shear"
        stretch.name = "stretch"
        stagger.name = "stagger"
        buckle.name = "buckle"
        propel.name = "propel"
        opening.name = "opening"

        # get correlation between neighboring basepairs among all helical parameters
        results = {}
        datasets = [shear, stretch, stagger, buckle, propel, opening]
        for ser1, ser2 in product(datasets, datasets):
            ser2_shifted = ser2.shift(axis=1)
            ser2_shifted[labels[0]] = ser2[labels[-1]]
            if ser1.name in constants.hp_angular and ser2.name in constants.hp_angular:
                method = self.circular
            elif (
                ser1.name in constants.hp_angular and ser2.name not in constants.hp_angular
            ) or (
                ser2.name in constants.hp_angular and ser1.name not in constants.hp_angular
            ):
                method = self.circlineal
            else:
                method = "pearson"  # type: ignore
            corr_data = ser1.corrwith(ser2_shifted, method=method)
            corr_data.index = corr_index
            results[f"{ser1.name}/{ser2.name}"] = corr_data
        result_df = pd.DataFrame.from_dict(results)
        result_df.index = corr_index  # type: ignore

        # save csv data
        result_df.to_csv(self.stage_io_dict["out"]["output_csv_path"])

        # create heatmap
        cmap = plt.get_cmap("bwr").copy()
        bounds = [-1, -0.8, -0.6, -0.4, -0.2, 0.2, 0.4, 0.6, 0.8, 1]
        num = cmap.N
        norm = mpl.colors.BoundaryNorm(bounds, num)  # type: ignore
        cmap.set_bad(color="gainsboro")
        fig, ax = plt.subplots(1, 1, dpi=300, figsize=(7.5, 5), tight_layout=True)
        im = ax.imshow(result_df, cmap=cmap, norm=norm, aspect="auto")
        plt.colorbar(im, ticks=[-1, -0.8, -0.6, -0.4, -0.2, 0.2, 0.4, 0.6, 0.8, 1])

        # axes
        xlocs = np.arange(len(result_df.columns))
        _ = ax.set_xticks(xlocs)
        _ = ax.set_xticklabels(result_df.columns.to_list(), rotation=90)

        ylocs = np.arange(len(result_df.index))
        _ = ax.set_yticks(ylocs)
        _ = ax.set_yticklabels(result_df.index.to_list())  # type: ignore

        ax.set_title(
            "Correlation for neighboring basepairs " "and pairs of helical parameters"
        )

        fig.tight_layout()
        fig.savefig(self.stage_io_dict["out"]["output_jpg_path"], format="jpg")
        plt.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0

    @staticmethod
    def circular(x1, x2):
        x1 = x1 * np.pi / 180
        x2 = x2 * np.pi / 180
        diff_1 = np.sin(x1 - x1.mean())
        diff_2 = np.sin(x2 - x2.mean())
        num = (diff_1 * diff_2).sum()
        den = np.sqrt((diff_1**2).sum() * (diff_2**2).sum())
        return num / den

    @staticmethod
    def circlineal(x1, x2):
        x2 = x2 * np.pi / 180
        rc = np.corrcoef(x1, np.cos(x2))[1, 0]
        rs = np.corrcoef(x1, np.sin(x2))[1, 0]
        rcs = np.corrcoef(np.sin(x2), np.cos(x2))[1, 0]
        num = (rc**2) + (rs**2) - 2 * rc * rs * rcs
        den = 1 - (rcs**2)
        correlation = np.sqrt(num / den)
        if np.corrcoef(x1, x2)[1, 0] < 0:
            correlation *= -1
        return correlation


def intrabpcorr(
    input_filename_shear: str,
    input_filename_stretch: str,
    input_filename_stagger: str,
    input_filename_buckle: str,
    input_filename_propel: str,
    input_filename_opening: str,
    output_csv_path: str,
    output_jpg_path: str,
    properties: Optional[dict] = None,
    **kwargs,
) -> int:
    """Create :class:`HelParCorrelation <intrabp_correlations.intrabpcorr.IntraBasePairCorrelation>` class and
    execute the :meth:`launch() <intrabp_correlations.intrabpcorr.IntraBasePairCorrelation.launch>` method."""
    return IntraBasePairCorrelation(**dict(locals())).launch()


intrabpcorr.__doc__ = IntraBasePairCorrelation.__doc__
main = IntraBasePairCorrelation.get_main(intrabpcorr, "Load .ser file from Canal output and calculate correlation between base pairs of the corresponding sequence.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the Curves class and the command line interface."""
import os
import zipfile
from typing import Optional
import shutil
from pathlib import Path
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class Curves(BiobbObject):
    """
    | biobb_dna Curves
    | Wrapper for the Cur+ executable  that is part of the Curves+ software suite.
    | The Cur+ program is used to analyze the structure of nucleic acids and their complexes.

    Args:
        input_struc_path (str): Trajectory or PDB input file. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/structure.stripped.trj>`_. Accepted formats: trj (edam:format_3910), pdb (edam:format_1476), netcdf (edam:format_3650), nc (edam:format_3650).
        input_top_path (str) (Optional): Topology file, needed along with .trj file (optional). File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/curvesplus/structure.stripped.top>`_. Accepted formats: top (edam:format_3881), pdb (edam:format_1476).
        output_cda_path (str): Filename for Curves+ output .cda file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/curves_trj_output.cda>`_. Accepted formats: cda (edam:format_2330).
        output_lis_path (str): Filename for Curves+ output .lis file. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/curves_trj_output.lis>`_. Accepted formats: lis (edam:format_2330).
        output_zip_path (str) (Optional): Filename for .zip files containing Curves+ output that is not .cda or .lis files. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/curvesplus/curves_trj_output.zip>`_. Accepted formats: zip (edam:format_3987).
        properties (dict):
            * **s1range** (*str*) - (None) Range of first strand. Must be specified in the form "start:end".
            * **s2range** (*str*) - (None) Range of second strand. Must be specified in the form "start:end".
            * **stdlib_path** (*str*) - ('standard') Path to Curves' standard library files for nucleotides. If not specified will look for 'standard' files in current directory.
            * **itst** (*int*) - (0) Iteration start index.
            * **itnd** (*int*) - (0) Iteration end index.
            * **itdel** (*int*) - (1) Iteration delimiter.
            * **ions** (*bool*) - (False) If True, helicoidal analysis of ions (or solvent molecules) around solute is carried out.
            * **test** (*bool*) - (False) If True, provide addition output in .lis file on fitting and axis generation.
            * **line** (*bool*) - (False) if True, find the best linear helical axis.
            * **fit** (*bool*) - (True) if True, fit a standard bases to the input coordinates (important for MD snapshots to avoid base distortions leading to noisy helical parameters).
            * **axfrm** (*bool*) - (False) if True, generates closely spaced helical axis frames as input for Canal and Canion.
            * **binary_path** (*str*) - (Cur+) Path to Curves+ executable, otherwise the program wil look for Cur+ executable in the binaries folder.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.curvesplus.biobb_curves import biobb_curves
            prop = {
                's1range': '1:12',
                's2range': '24:13',
            }
            biobb_curves(
                input_struc_path='/path/to/structure/file.trj',
                input_top_path='/path/to/topology/file.top',
                output_cda_path='/path/to/output/file.cda',
                output_lis_path='/path/to/output/file.lis',
                properties=prop)
    Info:
        * wrapped_software:
            * name: Curves
            * version: >=2.6
            * license: BSD 3-Clause
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
            self, input_struc_path, output_lis_path,
            output_cda_path, output_zip_path=None,
            input_top_path=None, properties=None, **kwargs) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            'in': {
                'input_struc_path': input_struc_path,
                'input_top_path': input_top_path
            },
            'out': {
                'output_lis_path': output_lis_path,
                'output_cda_path': output_cda_path,
                'output_zip_path': output_zip_path
            }
        }

        # Properties specific for BB
        self.s1range = properties.get('s1range', None)
        self.binary_path = properties.get('binary_path', 'Cur+')
        self.stdlib_path = properties.get('stdlib_path', None)
        self.s2range = properties.get('s2range', None)
        self.itst = properties.get('itst', 0)
        self.itnd = properties.get('itnd', 0)
        self.itdel = properties.get('itdel', 1)
        self.ions = ".t." if properties.get('ions', False) else ".f."
        self.test = ".t." if properties.get('test', False) else ".f."
        self.line = ".t." if properties.get('line', False) else ".f."
        self.fit = ".t." if properties.get('fit', True) else ".f."
        self.axfrm = ".t." if properties.get('axfrm', False) else ".f."
        self.properties = properties

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    def create_curvesplus_folder(self):
        """Create .curvesplus folder in the current temporal folder and copy the lib files inside."""
        # Create .curvesplus directory in temporary folder
        dst_dir = self.stage_io_dict.get("unique_dir", "") + '/.curvesplus'
        os.makedirs(dst_dir, exist_ok=True)
        # Get lib files from stdlib_path
        lib_files = list(Path(os.path.dirname(self.stdlib_path)).glob("*.lib"))
        # Copy each lib file to the .curvesplus directory in temporary folder
        for file in lib_files:
            shutil.copy(file, dst_dir)

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`Curves <biobb_dna.curvesplus.biobb_curves.Curves>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        if self.s1range is None:
            raise ValueError("property 's1range' must be specified!")
        if self.s2range is None:
            # compute s2range if not provided
            range1_end = int(self.s1range.split(":")[1])
            s2start = range1_end + 1
            s2end = 2 * range1_end
            self.s2range = f"{s2end}:{s2start}"

        # check standard library files location if not provided
        if self.stdlib_path is None:
            if os.getenv("CONDA_PREFIX", False):
                curves_aux_path = Path(
                    os.getenv("CONDA_PREFIX", "")) / ".curvesplus"
                # check if .curvesplus directory is in $CONDA_PREFIX
                if curves_aux_path.exists():
                    if len(list(curves_aux_path.glob("standard_*.lib"))) != 3:
                        raise FileNotFoundError(
                            "One or all standard library files "
                            f"missing from {curves_aux_path}! "
                            "Check files standard_b.lib, "
                            "standard_s.lib and standard_i.lib exist.")
                    self.stdlib_path = curves_aux_path / "standard"
                else:
                    raise FileNotFoundError(
                        ".curvesplus directory not found in "
                        f"{os.getenv('CONDA_PREFIX')} !"
                        "Please indicate where standard_*.lib files are "
                        "located with the stdlib_path property.")
                # copy standard library files to temporary folder
                shutil.copytree(curves_aux_path, self.stage_io_dict.get("unique_dir", "") + '/.curvesplus')
                relative_lib_path = '.curvesplus/standard'
            else:
                # CONDA_PREFIX undefined
                fu.log('CONDA_PREFIX undefined, please put the standard_b.lib, standard_s.lib and standard_i.lib files in the current working directory', self.out_log)
                self.stdlib_path = Path.cwd() / "standard"
                # create .curvesplus folder in the current temporal folder and copy the lib files inside
                self.create_curvesplus_folder()
                # set relative path
                relative_lib_path = '.curvesplus/standard'
        else:
            # create .curvesplus folder in the current temporal folder and copy the lib files inside
            self.create_curvesplus_folder()
            # set relative path
            path_parts = str(self.stdlib_path).split(os.sep)
            relative_lib_path = '.curvesplus/' + os.sep.join(path_parts[-1:])

        # change directory to temporary folder
        original_directory = os.getcwd()
        os.chdir(self.stage_io_dict.get("unique_dir", ""))

        # define temporary file names
        tmp_struc_input = Path(self.stage_io_dict['in']['input_struc_path']).name
        if self.stage_io_dict['in']['input_top_path'] is not None:
            tmp_top_input = Path(self.stage_io_dict['in']['input_top_path']).name

        # create intructions
        instructions = [
            f"{self.binary_path} <<! ",
            "&inp",
            f"  file={tmp_struc_input},"]
        if self.stage_io_dict['in']['input_top_path'] is not None:
            # add topology file if needed
            fu.log('Appending provided topology to command',
                   self.out_log, self.global_log)
            instructions.append(
                f"  ftop={tmp_top_input},")

        # create intructions
        instructions = instructions + [
            "  lis='curves_output',",
            f"  lib={relative_lib_path},",
            f"  ions={self.ions},",
            f"  test={self.test},",
            f"  line={self.line},",
            f"  fit={self.fit},",
            f"  axfrm={self.axfrm},",
            f"  itst={self.itst},itnd={self.itnd},itdel={self.itdel},",
            "&end",
            "2 1 -1 0 0",
            f"{self.s1range}",
            f"{self.s2range}",
            "!"
        ]
        self.cmd = ["\n".join(instructions)]
        fu.log('Creating command line with instructions and required arguments',
               self.out_log, self.global_log)

        # Run Biobb block
        self.run_biobb()

        # change back to original directory
        os.chdir(original_directory)

        # create zipfile and write output inside
        if self.stage_io_dict.get("out", {}).get("output_zip_path") is not None:
            zf = zipfile.ZipFile(
                Path(self.stage_io_dict["out"]["output_zip_path"]),
                "w")
            for curves_outfile in Path(self.stage_io_dict.get("unique_dir", "")).glob("curves_output*"):
                if curves_outfile.suffix not in (".cda", ".lis", ".zip"):
                    zf.write(
                        curves_outfile,
                        arcname=curves_outfile.name)
            zf.close()

        # rename cda and lis files
        (Path(self.stage_io_dict.get("unique_dir", "")) / "curves_output.cda").rename(
            self.stage_io_dict["out"]["output_cda_path"])
        (Path(self.stage_io_dict.get("unique_dir", "")) / "curves_output.lis").rename(
            self.stage_io_dict["out"]["output_lis_path"])

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return self.return_code


def biobb_curves(
        input_struc_path: str, output_lis_path: str, output_cda_path: str,
        input_top_path: Optional[str] = None, output_zip_path: Optional[str] = None,
        properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`Curves <biobb_dna.curvesplus.biobb_curves.Curves>` class and
    execute the :meth:`launch() <biobb_dna.curvesplus.biobb_curves.Curves.launch>` method."""
    return Curves(**dict(locals())).launch()


biobb_curves.__doc__ = Curves.__doc__
main = Curves.get_main(biobb_curves, "Execute Cur+ form the Curves+ software suite.")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Module containing the HelParTimeSeries class and the command line interface."""

import re
import zipfile
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import pandas as pd
from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger

from biobb_dna.utils import constants
from biobb_dna.utils.common import _from_string_to_list
from biobb_dna.utils.loader import read_series


class HelParTimeSeries(BiobbObject):
    """
    | biobb_dna HelParTimeSeries
    | Created time series and histogram plots for each base pair from a helical parameter series file.
    | The helical parameter series file is expected to be a table, with the first column being an index and the rest the helical parameter values for each base/basepair.

    Args:
        input_ser_path (str): Path to .ser file for helical parameter. File is expected to be a table, with the first column being an index and the rest the helical parameter values for each base/basepair. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/data/dna/canal_output_shift.ser>`_. Accepted formats: ser (edam:format_2330).
        output_zip_path (str): Path to output .zip files where data is saved. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_dna/master/biobb_dna/test/reference/dna/timeseries_output.zip>`_. Accepted formats: zip (edam:format_3987).
        properties (dict):
            * **sequence** (*str*) - (None) Nucleic acid sequence corresponding to the input .ser file. Length of sequence is expected to be the same as the total number of columns in the .ser file, minus the index column (even if later on a subset of columns is selected with the *usecols* option).
            * **bins** (*int*) - (None) Bins for histogram. Parameter has same options as matplotlib.pyplot.hist.
            * **helpar_name** (*str*) - (None) Helical parameter name. It must match the name of the helical parameter in the .ser input file. Values: majd, majw, mind, minw, inclin, tip, xdisp, ydisp, shear, stretch, stagger, buckle, propel, opening, rise, roll, twist, shift, slide, tilt, alphaC, alphaW, betaC, betaW, gammaC, gammaW, deltaC, deltaW, epsilC, epsilW, zetaC, zetaW, chiC, chiW, phaseC, phaseW.
            * **stride** (*int*) - (1000) granularity of the number of snapshots for plotting time series.
            * **seqpos** (*list*) - (None) list of sequence positions (columns indices starting by 1) to analyze.  If not specified it will analyse the complete sequence.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_dna.dna.dna_timeseries import dna_timeseries

            prop = {
                'helpar_name': 'twist',
                'seqpos': [1,2,3,4,5],
                'sequence': 'GCAACGTGCTATGGAAGC',
            }
            dna_timeseries(
                input_ser_path='/path/to/twist.ser',
                output_zip_path='/path/to/output/file.zip'
                properties=prop)
    Info:
        * wrapped_software:
            * name: In house
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl

    """

    def __init__(
        self, input_ser_path, output_zip_path, properties=None, **kwargs
    ) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {
                "input_ser_path": input_ser_path,
            },
            "out": {"output_zip_path": output_zip_path},
        }

        self.properties = properties
        self.sequence = properties.get("sequence", None)
        self.bins = properties.get("bins", "auto")
        self.stride = properties.get("stride", 10)
        self.seqpos = [
            int(elem) for elem in _from_string_to_list(properties.get("seqpos", None))
        ]
        self.helpar_name = properties.get("helpar_name", None)

        # get helical parameter from filename if not specified
        if self.helpar_name is None:
            for hp in constants.helical_parameters:
                if hp.lower() in Path(input_ser_path).name.lower():
                    self.helpar_name = hp
            if self.helpar_name is None:
                raise ValueError(
                    "Helical parameter name can't be inferred from file, "
                    "so it must be specified!"
                )
        else:
            if self.helpar_name not in constants.helical_parameters:
                raise ValueError(
                    "Helical parameter name is invalid! "
                    f"Options: {constants.helical_parameters}"
                )

        # get base length from helical parameter name
        if self.helpar_name.lower() in constants.hp_singlebases:
            self.baselen = 0
        else:
            self.baselen = 1
        # get unit from helical parameter name
        if self.helpar_name in constants.hp_angular:
            self.hp_unit = "Degrees"
        else:
            self.hp_unit = "Angstroms"

        # Check the properties
        self.check_properties(properties)
        self.check_arguments()

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`HelParTimeSeries <dna.dna_timeseries.HelParTimeSeries>` object."""

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        # check sequence
        if self.sequence is None or len(self.sequence) < 2:
            raise ValueError("sequence is null or too short!")

        # calculate cols with 0 index
        if self.seqpos:
            cols = [i - 1 for i in self.seqpos]
        else:
            cols = list(range(len(self.sequence)))

        # sort cols in ascending order
        cols.sort()

        # check seqpos for base pairs
        if self.seqpos and self.helpar_name in constants.hp_basepairs:
            if (max(cols) > len(self.sequence) - 2) or (min(cols) < 0):
                raise ValueError(
                    f"seqpos values must be between 1 and {len(self.sequence) - 1}"
                )
            if not (isinstance(self.seqpos, list) and len(self.seqpos) > 1):
                raise ValueError("seqpos must be a list of at least two integers")
        # check seqpos for non base pairs
        elif self.seqpos and self.helpar_name not in constants.hp_basepairs:
            if (max(cols) > len(self.sequence) - 1) or (min(cols) < 0):
                raise ValueError(
                    f"seqpos values must be between 1 and {len(self.sequence)}"
                )
            if not (isinstance(self.seqpos, list) and len(self.seqpos) > 1):
                raise ValueError("seqpos must be a list of at least two integers")

        if self.helpar_name in constants.hp_basepairs:
            # remove first and last base pairs from cols if they match 0 and len(sequence)
            if min(cols) == 0:
                cols.pop(0)
            if max(cols) == len(self.sequence) - 1:
                cols.pop(-1)

            # discard first and last base(pairs) from sequence
            sequence = self.sequence[1:-1]
            # create indices list
            indices = cols.copy()
            # create subunits list from cols
            subunits = [f"{i+1}_{sequence[i-1:i+self.baselen]}" for i in cols]
            # clean subunits (leave only basepairs)
            pattern = re.compile(r"\d+_[A-Za-z]{2}")
            # get removed items
            removed_items = [s for s in subunits if not pattern.fullmatch(s)]
            # get indices of removed items (in integer format and starting from 0)
            removed_numbers = [
                int(match.group())
                for item in removed_items
                if (match := re.match(r"\d+", item))
            ]
            removed_numbers = list(map(int, removed_numbers))
            removed_numbers = [int(i) - 1 for i in removed_numbers]
            # remove non basepairs from subunits and indices
            subunits = [s for s in subunits if pattern.fullmatch(s)]
            indices = [i for i in indices if i not in removed_numbers]
        else:
            sequence = self.sequence
            # create indices list
            indices = cols.copy()
            # trick for getting the index column from the .ser file
            indices.insert(0, 0)
            # create subunits list from cols
            subunits = [f"{i+1}_{sequence[i:i+1+self.baselen]}" for i in cols]

        # read input .ser file
        ser_data = read_series(
            self.stage_io_dict["in"]["input_ser_path"], usecols=indices
        )

        # get columns for selected bases
        ser_data.columns = subunits

        # write output files for all selected bases (one per column)
        zf = zipfile.ZipFile(Path(self.stage_io_dict["out"]["output_zip_path"]), "w")
        for col in ser_data.columns:
            # unstack columns to prevent errors from repeated base pairs
            column_data = ser_data[[col]].unstack().dropna().reset_index(drop=True)
            column_data.name = col
            fu.log(f"Computing base number {col}...")

            # column series
            series_colfn = f"series_{self.helpar_name}_{col}"
            column_data.to_csv(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{series_colfn}.csv"
            )
            # save table
            zf.write(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{series_colfn}.csv",
                arcname=f"{series_colfn}.csv",
            )

            fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
            reduced_data = column_data.iloc[:: self.stride]
            axs.plot(reduced_data.index, reduced_data.to_numpy())
            axs.set_xlabel("Time (Snapshots)")
            axs.set_ylabel(f"{self.helpar_name.capitalize()} ({self.hp_unit})")
            axs.set_title(
                f"Helical Parameter vs Time: {self.helpar_name.capitalize()} "
                "(base pair "
                f"{'step' if self.baselen == 1 else ''} {col})"
            )
            fig.savefig(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{series_colfn}.jpg",
                format="jpg",
            )
            # save plot
            zf.write(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{series_colfn}.jpg",
                arcname=f"{series_colfn}.jpg",
            )
            plt.close()

            # columns histogram
            hist_colfn = f"hist_{self.helpar_name}_{col}"
            fig, axs = plt.subplots(1, 1, dpi=300, tight_layout=True)
            ybins, x, _ = axs.hist(column_data, bins=self.bins)
            pd.DataFrame({self.helpar_name: x[:-1], "density": ybins}).to_csv(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{hist_colfn}.csv",
                index=False,
            )
            # save table
            zf.write(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{hist_colfn}.csv",
                arcname=f"{hist_colfn}.csv",
            )

            axs.set_ylabel("Density")
            axs.set_xlabel(f"{self.helpar_name.capitalize()} ({self.hp_unit})")
            fig.savefig(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{hist_colfn}.jpg",
                format="jpg",
            )
            # save plot
            zf.write(
                Path(self.stage_io_dict.get("unique_dir", "")) / f"{hist_colfn}.jpg",
                arcname=f"{hist_colfn}.jpg",
            )
            plt.close()
        zf.close()

        # Copy files to host
        self.copy_to_host()

        # Remove temporary file(s)
        self.remove_tmp_files()

        self.check_arguments(output_files_created=True, raise_exception=False)

        return 0


def dna_timeseries(
    input_ser_path: str,
    output_zip_path: str,
    properties: Optional[dict] = None,
    **kwargs,
) -> int:
    """Create :class:`HelParTimeSeries <dna.dna_timeseries.HelParTimeSeries>` class and
    execute the :meth:`launch() <dna.dna_timeseries.HelParTimeSeries.launch>` method."""
    return HelParTimeSeries(**dict(locals())).launch()


dna_timeseries.__doc__ = HelParTimeSeries.__doc__
main = HelParTimeSeries.get_main(dna_timeseries, "Created time series and histogram plots for each base pair from a helical parameter series file.")

if __name__ == '__main__':
    main()
